import numpy as np
from matplotlib import pyplot as plt
import astropy.units as u
from pyspeckit.spectrum.units import SpectroscopicAxis
import warnings


# =======================================================================================================================

class Plotter(object):
    def __init__(self, ucube, fittype, ncomp_list=None, spec_unit='km/s', bunit=None):
        """
        Initialize the Plotter class.

        Parameters
        ----------
        ucube : UltraCube
            The UltraCube object containing the spectral cube and parameter cubes.
        fittype : str
            The type of model fit to use ('nh3_multi_v' or 'n2hp_multi_v').
        ncomp_list : list of int, optional
            List of component numbers to plot. If None, all components are used.
        spec_unit : str, optional
            The spectral unit for the cube. Default is 'km/s'.
        bunit : str or Quantity, optional
            The desired unit for the cube's data (e.g., 'K' for brightness temperature or 'Jy' for flux).
        """
        self.ucube = ucube
        self.cube = self.ucube.cube.with_spectral_unit(spec_unit, velocity_convention='radio')
        self.xarr = SpectroscopicAxis(self.cube.spectral_axis.value,
                                      unit=spec_unit,
                                      refX=self.cube._header['RESTFRQ'],
                                      velocity_convention='radio')

        # Convert bunit to a Unit if provided as a string
        if isinstance(bunit, str):
            bunit = u.Unit(bunit)

        # Set cube's unit and y-axis label
        if bunit is not None:
            if hasattr(self.cube, 'unit') and (self.cube.unit is not None and self.cube.unit != ''):
                try:
                    # Attempt to convert the cube's unit to bunit
                    self.cube = self.cube.to(bunit)
                    self.ylab = f"Intensity ({bunit.to_string()})"
                except u.UnitConversionError:
                    warnings.warn(f"Incompatible units: cube's unit ({self.cube.unit}) cannot be converted to {bunit}.")
                    self.ylab = f"Intensity ({self.cube.unit.to_string()})"
            else:
                # Cube has no unit; assume bunit as the unit and set the label accordingly
                warnings.warn("Cube has no unit attribute; setting y-axis label to specified bunit.")
                self.ylab = f"Intensity ({bunit.to_string()})"
        else:
            # No bunit specified; set y label based on cube's unit if available
            if hasattr(self.cube, 'unit') and (self.cube.unit is not None and self.cube.unit != ''):
                if self.cube.unit.is_equivalent(u.K):
                    self.ylab = r"$T_{\mathrm{MB}}$ (K)"
                elif self.cube.unit.is_equivalent(u.Jy):
                    self.ylab = r"Flux Density (Jy)"
                elif self.cube.unit.is_equivalent(u.mJy):
                    self.ylab = r"Flux Density (mJy)"
                else:
                    # Generic label if unit is unknown or unsupported
                    self.ylab = f"Intensity ({self.cube.unit.to_string()})"
            else:
                # No unit in cube and no bunit specified, set to generic label
                warnings.warn("Cube data has no unit attribute and no bunit specified; setting y-axis label to 'Intensity'.")
                self.ylab = "Intensity"

        # Set the x-axis label
        self.xlab = r"$v_{\mathrm{LSR}}$ (km s$^{-1}$)"

        # Handle the fittype selection
        if fittype == "nh3_multi_v":
            from ..spec_models.ammonia_multiv import ammonia_multi_v
            self.model_func = ammonia_multi_v
        elif fittype == "n2hp_multi_v":
            from ..spec_models.n2hp_multiv import n2hp_multi_v
            self.model_func = n2hp_multi_v
        else:
            raise ValueError(f"{fittype} is not one of the currently accepted fittypes.")

        # Process ncomp_list for parcubes
        self.parcubes = {}
        if ncomp_list is None:
            for key in self.ucube.pcubes:
                self.parcubes[key] = self.ucube.pcubes[key].parcube
        else:
            for n in ncomp_list:
                self.parcubes[str(n)] = self.ucube.pcubes[str(n)].parcube



    def plot_spec_grid(self, x, y, size=3, xsize=None, ysize=None, xlim=None, ylim=None, figsize=None, **kwargs):
        """
        Plot a grid of spectra centered at (x, y).

        Parameters
        ----------
        x : int
            X-coordinate of the central pixel.
        y : int
            Y-coordinate of the central pixel.
        size : int, optional
            Size of the grid (must be odd). Default is 3.
        xsize : int, optional
            Number of columns in the grid. Default is size.
        ysize : int, optional
            Number of rows in the grid. Default is size.
        xlim : tuple, optional
            X-axis limits for the plot, in their native units.
        ylim : tuple, optional
            Y-axis limits for the plot, in their native units.
        figsize : tuple, optional
            Size of the figure.
        **kwargs : dict
            Additional keyword arguments passed to `plot_spec_grid`.
        """
        # add defaults that can be superseded
        kwargs = {'xlab': self.xlab, 'ylab': self.ylab, **kwargs}

        self.fig, self.axs = \
            plot_spec_grid(self.cube, x, y, size=size, xsize=xsize, ysize=ysize, xlim=xlim, ylim=ylim, figsize=figsize, **kwargs)


    def plot_spec(self, x, y, ax=None, xlab=None, ylab=None, **kwargs):
        """
        Plot a single spectrum at (x, y).

        Parameters
        ----------
        x : int
            X-coordinate of the pixel.
        y : int
            Y-coordinate of the pixel.
        ax : matplotlib.axes.Axes, optional
            Axes object to plot on. If None, a new figure and axes are created.
        xlab : str, optional
            X-axis label. Default is the LSR velocity label.
        ylab : str, optional
            Y-axis label. Default is the main beam temperature label.
        **kwargs : dict
            Additional keyword arguments passed to `plot_spec`.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The figure object.
        ax : matplotlib.axes.Axes
            The axes object.
        """
        spc = self.cube[:, y, x]
        if xlab is None:
            xlab = self.xlab
        if ylab is None:
            ylab = self.ylab
        self.fig, self.axs = plot_spec(spc, xarr=self.xarr, ax=ax, xlab=xlab, ylab=ylab, **kwargs)
        return self.fig, self.axs


    def plot_fit(self, x, y, ax=None, ncomp=None, **kwargs):
        """
        Plot a model fit for a spectrum at (x, y).

        Parameters
        ----------
        x : int
            X-coordinate of the pixel.
        y : int
            Y-coordinate of the pixel.
        ax : matplotlib.axes.Axes, optional
            Axes object to plot on. If None, a new figure and axes are created.
        ncomp : int
            The component number to plot.
        **kwargs : dict
            Additional keyword arguments passed to `plot_model`.
        """
        if ax is None:
            self.fig, ax = plt.subplots()
            self.axs = ax
        plot_model(self.parcubes[str(ncomp)][:, y, x], self.model_func, self.xarr, ax, ncomp, **kwargs)


    def plot_fits_grid(self, x, y, ncomp, size=3, xsize=None, ysize=None, xlim=None, ylim=None,
                       figsize=None, origin='lower', mod_all=True, savename=None, **kwargs):
        """
        Plot a grid of model fits centered at (x, y).

        Parameters
        ----------
        x : int
            X-coordinate of the central pixel.
        y : int
            Y-coordinate of the central pixel.
        ncomp : int
            The component number to plot.
        size : int, optional
            Size of the grid (must be odd). Default is 3.
        xsize : int, optional
            Number of columns in the grid. Default is size.
        ysize : int, optional
            Number of rows in the grid. Default is size.
        xlim : tuple, optional
            X-axis limits for the plot.
        ylim : tuple, optional
            Y-axis limits for the plot.
        figsize : tuple, optional
            Size of the figure.
        origin : {'lower', 'upper'}, optional
            Origin of the grid. Default is 'lower'.
        mod_all : bool, optional
            Whether to plot all model components. Default is True.
        savename : str, optional
            If provided, save the figure to the given filename.
        **kwargs : dict
            Additional keyword arguments passed to `plot_fits_grid`.
        """
        # add defaults that can be superseded
        kwargs = {'xlab': self.xlab, 'ylab': self.ylab, **kwargs}

        self.fig, self.axs = \
            plot_fits_grid(self.cube, self.parcubes[str(ncomp)], self.model_func, x, y, self.xarr,
                           ncomp=ncomp, size=size, xsize=xsize, ysize=ysize, xlim=xlim, ylim=ylim,
                           figsize=figsize, origin=origin, mod_all=mod_all, savename=savename,
                           **kwargs)

# =======================================================================================================================
# None-plotting functions

def get_cube_slab(cube, vZoomLims=(-5, 20)):
    """
    Extract a spectral slab from the cube over the specified velocity range.

    Parameters
    ----------
    cube : SpectralCube
        The spectral cube to extract the slab from.
    vZoomLims : tuple of float, optional
        The velocity limits for the slab, in km/s. Default is (-5, 20).

    Returns
    -------
    cube_s : SpectralCube
        The extracted spectral slab.
    xarr : SpectroscopicAxis
        The corresponding spectroscopic axis for the slab.
    """
    # currently incomplete. it will be used to save some computation time when calculating the model and plotting
    cube = cube.with_spectral_unit(u.km / u.s, velocity_convention='radio')
    cube_s = cube.spectral_slab(vZoomLims[0] * u.km / u.s, vZoomLims[1] * u.km / u.s)

    # SpectroscopicAxis has the advantage of being able to performed unit conversion automatically
    xarr = SpectroscopicAxis(cube_s.spectral_axis.value, unit=cube_s.spectral_axis.unit,
                             refX=cube_s._header['RESTFRQ'], velocity_convention='radio')
    return cube_s, xarr

# =======================================================================================================================
# None-class functions

def plot_spec(spc, xarr=None, ax=None, fill=False, xlab=None, ylab=None, **kwargs):
    """
    Plot a spectrum.

    Parameters
    ----------
    spc : Spectrum
        The spectrum to plot.
    xarr : SpectroscopicAxis, optional
        The spectroscopic axis to use for the x-axis. If None, the spectrum's spectral axis is used.
    ax : matplotlib.axes.Axes, optional
        Axes object to plot on. If None, a new figure and axes are created.
    fill : bool, optional
        Whether to fill the area under the curve. Default is False.
    xlab : str, optional
        X-axis label.
    ylab : str, optional
        Y-axis label.
    **kwargs : dict
        Additional keyword arguments passed to `ax.plot` or `ax.fill_between`.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object, if `ax` is None.
    ax : matplotlib.axes.Axes
        The axes object.
    """
    if 'c' in kwargs:
        if 'color' in kwargs:
            raise TypeError("Got both 'color' and 'c', which are aliases of one another")
        else:
            kwargs['color'] = kwargs['c']
            del kwargs['c']
    elif 'color' not in kwargs:
        kwargs['color'] = '0.65'

    kwargs_df = dict(lw=1)  # default kwargs
    kwargs = {**kwargs_df, **kwargs}

    return_fig = False
    if ax is None:
        fig = plt.figure()
        ax = fig.gca()
        return_fig = True

    if xarr is None:
        xarr = spc.spectral_axis

    if fill:
        ax.fill_between(xarr.value, 0, spc, **kwargs)
    else:
        ax.plot(xarr.value, spc, **kwargs)

    if xlab is not None:
        ax.set_xlabel(xlab)
    if ylab is not None:
        ax.set_ylabel(ylab)

    if return_fig:
        return fig, ax


def get_spec_grid(size=3, xsize=None, ysize=None, figsize=None):
    """
    Create a grid of subplots for spectra.

    Parameters
    ----------
    size : int, optional
        Size of the grid (must be odd). Default is 3.
    xsize : int, optional
        Number of columns in the grid. Default is size.
    ysize : int, optional
        Number of rows in the grid. Default is size.
    figsize : tuple, optional
        Size of the figure.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object.
    axs : numpy.ndarray of matplotlib.axes.Axes
        The array of axes objects.
    """
    if size % 2 == 0 and size > 0:
        raise ValueError("Size provided must be an odd, positive integer that is not zero.")

    if xsize is None:
        xsize = size

    if ysize is None:
        ysize = size

    # initiate the plot grid
    fig, axs = plt.subplots(ysize, xsize, sharex='all', sharey='all', gridspec_kw={'hspace': 0, 'wspace': 0}, figsize=figsize)

    return fig, axs


def plot_spec_grid(cube, x, y, size=3, xsize=None, ysize=None, xlim=None, ylim=None, figsize=None,
                   origin='lower', **kwargs):
    """
    Plot a grid of spectra from the cube centered at (x, y).

    Parameters
    ----------
    cube : SpectralCube
        The spectral cube to plot.
    x : int
        X-coordinate of the central pixel.
    y : int
        Y-coordinate of the central pixel.
    size : int, optional
        Size of the grid (must be odd). Default is 3.
    xsize : int, optional
        Number of columns in the grid. Default is size.
    ysize : int, optional
        Number of rows in the grid. Default is size.
    xlim : tuple, optional
        X-axis limits for the plot.
    ylim : tuple, optional
        Y-axis limits for the plot.
    figsize : tuple, optional
        Size of the figure.
    origin : {'lower', 'upper'}, optional
        Origin of the grid. Default is 'lower'.
    **kwargs : dict
        Additional keyword arguments passed to `plot_spec`.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object.
    axs : numpy.ndarray of matplotlib.axes.Axes
        The array of axes objects.
    """
    # grab the x y labels and ensure it wasn't passed downstream
    if 'xlab' in kwargs:
        xlab = kwargs['xlab']
        del kwargs['xlab']

    if 'ylab' in kwargs:
        ylab = kwargs['ylab']
        del kwargs['ylab']

    fig, axs = get_spec_grid(size=size, xsize=xsize, ysize=ysize, figsize=figsize, **kwargs)
    ysize, xsize = axs.shape

    xpad = int(xsize / 2)
    ypad = int(ysize / 2)

    # Get a subcube
    scube = cube[:, y - ypad: y + ypad + 1, x - xpad: x + xpad + 1]

    if ylim is None:
        ymax = scube.max()
        ylim = (None, ymax * 1.1)

    # Ensure user-provided xlim and ylim have compatible units with data
    xlim = ensure_units_compatible(xlim, scube.spectral_axis.unit)
    ylim = ensure_units_compatible(ylim, scube.unit)

    # Strip units from xlim and ylim if they are Quantities
    xlim = strip_units(xlim)
    ylim = strip_units(ylim)

    # Existing code for plotting the spectra over the grid
    for index, ax in np.ndenumerate(axs):
        yi, xi = index
        if origin == 'lower':
            yi = ysize - 1 - yi
        elif origin != 'upper':
            raise KeyError(f"The keyword '{origin}' is unsupported.")

        spc = scube[:, yi, xi]
        plot_spec(spc, ax=ax, **kwargs)

    if origin == 'lower':
        axs = np.flip(axs, axis=0)

    for ax in axs.flat:
        ax.label_outer()

    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    # Provide a common labeling ax
    fig.add_subplot(111, frameon=False, zorder=-100)
    plt.tick_params(labelcolor='none', top=False, bottom=False, left=False, right=False)
    plt.grid(False)
    plt.xlabel(xlab)
    plt.ylabel(ylab)

    return fig, axs


def plot_model(para, model_func, xarr, ax, ncomp, **kwargs):
    """
    Plot a model fit for a spectrum.

    Parameters
    ----------
    para : numpy.ndarray
        The parameter array for the model.
    model_func : function
        The model function to use for the fit.
    xarr : SpectroscopicAxis
        The spectroscopic axis for the x-axis.
    ax : matplotlib.axes.Axes
        Axes object to plot on.
    ncomp : int
        The number of components in the model.
    **kwargs : dict
        Additional keyword arguments passed to `plot_spec`.
    """
    for i in range(ncomp):
        pp = para[i * 4:(i + 1) * 4]
        mod = model_func(xarr, *pp)
        if 'alpha' not in kwargs:
            kwargs['alpha'] = 0.6
        plot_spec(mod, xarr, ax, fill=True, color=f'C{i}', **kwargs)

    mod_tot = model_func(xarr, *para)
    plot_spec(mod_tot, xarr, ax, c='0.1', zorder=30, **kwargs)


def plot_fits_grid(cube, para, model_func, x, y, xarr, ncomp, size=3, xsize=None, ysize=None, xlim=None, ylim=None,
                   figsize=None, origin='lower', mod_all=True, savename=None, **kwargs):
    """
    Plot a grid of model fits from the cube centered at (x, y).

    Parameters
    ----------
    cube : SpectralCube
        The spectral cube to plot.
    para : numpy.ndarray
        The parameter array for the model.
    model_func : function
        The model function to use for the fit.
    x : int
        X-coordinate of the central pixel.
    y : int
        Y-coordinate of the central pixel.
    xarr : SpectroscopicAxis
        The spectroscopic axis for the x-axis.
    ncomp : int
        The number of components in the model.
    size : int, optional
        Size of the grid (must be odd). Default is 3.
    xsize : int, optional
        Number of columns in the grid. Default is size.
    ysize : int, optional
        Number of rows in the grid. Default is size.
    xlim : tuple, optional
        X-axis limits for the plot.
    ylim : tuple, optional
        Y-axis limits for the plot.
    figsize : tuple, optional
        Size of the figure.
    origin : {'lower', 'upper'}, optional
        Origin of the grid. Default is 'lower'.
    mod_all : bool, optional
        Whether to plot all model components. Default is True.
    savename : str, optional
        If provided, save the figure to the given filename.
    **kwargs : dict
        Additional keyword arguments passed to `plot_model`.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object.
    axs : numpy.ndarray of matplotlib.axes.Axes
        The array of axes objects.
    """
    fig, axs = plot_spec_grid(cube, x, y, size=size, xsize=xsize, ysize=ysize, xlim=xlim, ylim=ylim,
                              figsize=figsize, origin=origin, **kwargs)

    ysize, xsize = axs.shape
    xpad = int(xsize / 2)
    ypad = int(ysize / 2)

    if mod_all:
        for j in range(y - ypad, y + ypad + 1):
            for i in range(x - xpad, x + xpad + 1):
                plot_model(para[:, j, i], model_func, xarr, ax=axs[j - y + ypad, i - x + xpad], ncomp=ncomp, lw=1)
    else:
        # Plot the central pixel only
        plot_model(para[:, y, x], model_func, xarr, ax=axs[ypad, xpad], ncomp=ncomp, lw=1)

    if savename is not None:
        fig.savefig(savename, bbox_inches='tight')

    return fig, axs


#=======================

def strip_units(lim):
    """
    Helper function to strip units from a limit tuple if it contains Quantity.

    Parameters
    ----------
    lim : tuple or None
        A tuple of limits (min, max) which may contain Quantity instances with units.

    Returns
    -------
    tuple
        A tuple with units stripped, containing only numeric values.
    """
    if lim is not None:
        lim = tuple(val.value if isinstance(val, u.Quantity) else val for val in lim)
    return lim


def ensure_units_compatible(lim, data_unit, suppress_warning=False):
    """
    Ensure the limits have compatible units with the data, converting if needed.
    If the data has no units, strip units from the limits directly and issue a warning if desired.

    Parameters
    ----------
    lim : tuple or None
        The limit tuple (min, max) to be checked and potentially converted.
    data_unit : astropy.units.Unit or None
        The unit of the data to which the limits should be compatible.
    suppress_warning : bool, optional
        If True, suppresses warnings when stripping units from limits.

    Returns
    -------
    tuple
        Limits in compatible units with data, or stripped of units if data has no units.
    """
    if lim is not None:
        # Strip units if data has no units, and issue a warning if not suppressed
        if data_unit is None:
            for val in lim:
                if isinstance(val, u.Quantity) and not suppress_warning:
                    warnings.warn("Data has no units; stripping units from provided limit.")
            lim = strip_units(lim)
        else:
            # Convert each limit to match data units if it's a Quantity with a different unit
            lim = tuple(
                val.to(data_unit) if (val is not None and isinstance(val, u.Quantity) and val.unit != data_unit) else val
                for val in lim
            )
    return lim



"""
The `mufasa.guess_refine` module provides tools for cleaning, normalizing, and
interpolating fitted parameter maps and generating refined guesses for fitting.
"""

__author__ = 'mcychen'

#=======================================================================================================================
import numpy as np
import warnings

from astropy.stats import mad_std
from astropy.wcs import WCS
from skimage.morphology import remove_small_objects, binary_dilation, disk, remove_small_holes
from scipy.ndimage.filters import median_filter
from scipy.interpolate import CloughTocher2DInterpolator as intp
from scipy.interpolate import griddata
from .utils.fits_utils import get_pixel_mapping
from astropy.convolution import Gaussian2DKernel, convolve

from scipy.spatial.qhull import QhullError

from pyspeckit.spectrum.models.ammonia_constants import freq_dict
from astropy import units as u
nu0_nh3 = freq_dict['oneone'] * u.Hz
nu0_nh3 = nu0_nh3.to("GHz").value

from .utils import interpolate
from . import moment_guess as mmg
#=======================================================================================================================
from .utils.mufasa_log import get_logger
logger = get_logger(__name__)
#=======================================================================================================================


def quick_2comp_sort(data_cnv, filtsize=2, method="error_v", nu=nu0_nh3, f_tau=0.5):
    # use median filtered vlsr & sigma maps as a velocity reference to sort the two components
    # f_tau is the factor to down scale tau to minimic the effective tau of the main hyperfines

    if method == "error_v":
        # arange the maps so the component with the least vlsr errors is the first component
        swapmask = data_cnv[8] > data_cnv[12]
        data_cnv = mask_swap_2comp(data_cnv, swapmask)

    elif method == "Tpeak":
        # sort by the peak brigthness temperature using the tau & tax parameter
        # the brigther component is placed as the first component (the further away from the observer)
        Tb0_a = mmg.peakT(data_cnv[3], data_cnv[4]*f_tau, nu=nu)
        Tb0_b = mmg.peakT(data_cnv[6], data_cnv[7]*f_tau, nu=nu)
        swapmask = Tb0_b > Tb0_a
        data_cnv = mask_swap_2comp(data_cnv, swapmask)

    elif method == "tautex":
        # sort by the relative emission brightness proxy
        # the brigther component is placed as the first component (the further away from the observer)
        Tb0_a = data_cnv[3]*data_cnv[4]
        Tb0_b = data_cnv[6]*data_cnv[7]
        swapmask = Tb0_b > Tb0_a
        data_cnv = mask_swap_2comp(data_cnv, swapmask)


    elif method == "tau":
        # placing the optically thicker component in the back (1st component)
        # would only recommend if the degeneracy between tau and tex has been somewhat addressed already
        swapmask = data_cnv[7] > data_cnv[4]
        data_cnv = mask_swap_2comp(data_cnv, swapmask)


    elif method == "chen2020":
        # this is the method used by Chen+ 2020 ApJ

        # arange the maps so the component with the least vlsr errors is the first component
        swapmask = data_cnv[8] > data_cnv[12]
        data_cnv = mask_swap_2comp(data_cnv, swapmask)

        # the use the vlsr error in the first component as the reference and sort the component based on their similarities
        # to this reference (similary bright structures should have similar errors)
        ref = median_filter(data_cnv[8], size=(filtsize, filtsize))
        swapmask = np.abs(data_cnv[8] - ref) > np.abs(data_cnv[12] - ref)
        data_cnv = mask_swap_2comp(data_cnv, swapmask)

        def dist_metric(p1, p2):
            # use the first map (the one that should have the smallest error, hense more reliable) to compute
            #  distance metric based on their similarities to the median filtered quantity
            p_refa = median_filter(p1, size=(filtsize, filtsize))
            #p_refb = median_filter(p2, size=(filtsize, filtsize))

            # distance of the current arangment to the median
            del_pa = np.abs(p1 - p_refa)

            # distance of the swapped arangment to the median
            del_pb = np.abs(p2 - p_refa)
            return del_pa, del_pb

        dist_va, dist_vb = dist_metric(data_cnv[0], data_cnv[4])
        dist_siga, dist_sigb = dist_metric(data_cnv[1], data_cnv[5])

        # use both the vlsr and the sigma as a distance metric
        swapmask = np.hypot(dist_va, dist_siga) > np.hypot(dist_vb, dist_sigb)
        data_cnv= mask_swap_2comp(data_cnv, swapmask)

    return data_cnv


def mask_swap_2comp(data_cnv, swapmask):
    # swap data over the mask
    data_cnv= data_cnv.copy()
    data_cnv[0:4,swapmask], data_cnv[4:8,swapmask] = data_cnv[4:8,swapmask], data_cnv[0:4,swapmask]
    data_cnv[8:12,swapmask], data_cnv[12:16,swapmask] = data_cnv[12:16,swapmask], data_cnv[8:12,swapmask]
    return data_cnv


def guess_from_cnvpara(data_cnv, header_cnv, header_target, mask=None, tau_thresh=1, clean_map=True):
    # a wrapper to make guesses based on the parameters fitted to the convolved data
    npara = 4
    ncomp = int(data_cnv.shape[0]/npara/2)

    data_cnv = data_cnv.copy()
    data_cnv[data_cnv == 0] = np.nan

    # clean up the maps based on vlsr errors
    if ncomp == 1:
        std_thres = 3
    else:
        std_thres = 1

    if ncomp == 2:
        data_cnv = quick_2comp_sort(data_cnv, filtsize=2, method="error_v")

    if clean_map:
        data_cnv = simple_para_clean(data_cnv, ncomp, npara=npara, std_thres=std_thres)
    # remove the error component
    data_cnv = data_cnv[0:npara*ncomp]

    for i in range (0, ncomp):
        data_cnv[i*npara:i*npara+npara] = refine_each_comp(data_cnv[i*npara:i*npara+npara], mask, tau_thresh=tau_thresh)

    # regrid the guess back to that of the original data
    hdr_conv = get_celestial_hdr(header_cnv)
    hdr_final = get_celestial_hdr(header_target)

    guesses_final = []

    newmask = np.any(np.isfinite(data_cnv), axis=0)
    newmask = remove_small_holes(newmask, 25)

    # regrid the guesses
    for gss in data_cnv:
        if False:
            newmask = np.isfinite(gss)
            # removal holes with areas that smaller than a 5 by 5 square
            newmask = remove_small_holes(newmask, 25)
            # create a mask to regrid over
            newmask = regrid(newmask, hdr_conv, hdr_final, dmask=None, method='nearest')
            newmask = newmask.astype('bool')

            new_guess = regrid(gss, hdr_conv, hdr_final, dmask=newmask)

            # expand the interpolation a bit, since regridding can often miss some pixels due to aliasing
            newmask_l = binary_dilation(newmask)
            newmask_l = binary_dilation(newmask_l)

            kernel = Gaussian2DKernel(1)
            new_guess_cnv = convolve(new_guess, kernel, boundary='extend')
            new_guess_cnv[~newmask_l] = np.nan
            # retrain the originally interpolataed values within the original mask
            mask_finite = np.isfinite(new_guess)
            new_guess_cnv[mask_finite] = new_guess[mask_finite]
        else:
            new_guess_cnv = interpolate.iter_expand(gss, mask=newmask)
            new_guess_cnv = regrid(new_guess_cnv, hdr_conv, hdr_final, dmask=None)

            # insure the footprint wasn't made smaller due to regriding
            mask_l = np.isfinite(new_guess_cnv)
            mask_l = binary_dilation(mask_l, disk(2))
            new_guess_cnv = interpolate.iter_expand(new_guess_cnv, mask=mask_l)

        guesses_final.append(new_guess_cnv)

    return np.array(guesses_final)


def tautex_renorm(taumap, texmap, tau_thresh = 0.21, tex_thresh = 15.0, nu=nu0_nh3):

    # attempt to re-normalize the tau & text values at the optically thin regime (where the two are degenerate)
    # note, the latest recipe also works for the optically thick regime in principle
    # only emission with lower amplitude than TA_ltau_thres and tau < tau_thin will have tex > tex_thresh recalculated
    #  (i.e., expected to be optically thin

    f_tau = 0.5 # a factor to minic the effecitve tau of the main hyperfines

    isthin = np.logical_and(taumap < tau_thresh, np.isfinite(taumap))
    TA_lowtau = mmg.peakT(texmap[isthin], taumap[isthin]*f_tau, nu=nu)
    TA_ltau_thres = 0.5 # where tau ~1 for Tex = 3.5; tau with Ta above this diverges quickly
    # assume a fixed Tex for low TA
    tex_thin = 3.5      # note: at Tk = 30K, n = 1e3, N = 1e13, & sig = 0.2 km.s --> Tex = 3.49 K, tau = 0.8
    tau_thin = tau_thresh#1.0      # where the main hyperfines of NH3 (1,1) starts to get optically thick

    # for when tau is less than tau_thresh
    texmap[isthin] = mmg.get_tex(TA_lowtau, tau=tau_thresh*f_tau) #note: tex can be higher than at 40K at Ta~7K
    taumap[isthin] = tau_thresh

    # optically thin gas are also unlikely to have high spatial density and thus high Tex
    hightex = np.logical_and(texmap > tex_thresh, np.isfinite(texmap))
    TA_hightex = mmg.peakT(texmap[hightex], taumap[hightex]*f_tau, nu=nu)
    mask = TA_hightex < TA_ltau_thres # only renormalize high tex when Ta is less than the threshold
    mask = np.logical_and(mask, taumap[hightex] < tau_thin)

    texmap[hightex][mask] = tex_thin
    taumap[hightex][mask] = mmg.get_tau(TA_hightex[mask], tex=tex_thin, nu=nu)

    # note, tau values that are too low will be taken care of by refine_each_comp()
    return taumap, texmap


def refine_each_comp(guess_comp, mask=None, v_range=None, sig_range=None, tau_thresh=0.1):
    # refine guesses for each component, with values outside ranges specified below removed

    Tex_min = 3.0
    Tex_max = 8.0
    Tau_min = 0.2
    Tau_max = 8.0

    disksize = 1.0

    if mask is None:
        mask = master_mask(guess_comp)

    if v_range is None:
        vmin = None
        vmax = None
    else:
        vmin, vmax =v_range

    if sig_range is None:
        sigmin = None
        sigmax = None
    else:
        sigmin, sigmax = sig_range


    guess_comp[0] = refine_guess(guess_comp[0], min=vmin, max=vmax, mask=mask, disksize=disksize)
    guess_comp[1] = refine_guess(guess_comp[1], min=sigmin, max=sigmax, mask=mask, disksize=disksize)

    # re-normalize the degenerated tau & text for the purpose of estimate guesses
    guess_comp[3], guess_comp[2] = tautex_renorm(guess_comp[3], guess_comp[2], tau_thresh = tau_thresh)

    # place a more "strict" limits for Tex and Tau guessing than the fitting itself
    guess_comp[2] = refine_guess(guess_comp[2], min=Tex_min, max=Tex_max, mask=mask, disksize=disksize)
    guess_comp[3] = refine_guess(guess_comp[3], min=Tau_min, max=Tau_max, mask=mask, disksize=disksize)
    return guess_comp


def simple_para_clean(pmaps, ncomp, npara=4, std_thres = 2):
    # clean parameter maps based on their error values

    pmaps=pmaps.copy()

    # remove component with vlsrErr that is number of sigma off from the median as specified below

    pmaps[pmaps == 0] = np.nan

    # loop through each component
    for i in range (0, ncomp):
        # get the STD and Medians of the vlsr errors
        std_vErr = mad_std(pmaps[(i+ncomp)*npara][np.isfinite(pmaps[(i+ncomp)*npara])])
        median_vErr = np.median(pmaps[(i+ncomp)*npara][np.isfinite(pmaps[(i+ncomp)*npara])])

        # remove outlier pixels
        mask = pmaps[(i+ncomp)*npara] > median_vErr + std_vErr*std_thres

        pmaps[i*npara:(i+1)*npara, mask] = np.nan
        pmaps[(i+ncomp)*npara:(i+ncomp+1)*npara, mask] = np.nan

    return pmaps


def get_celestial_hdr(header):
    # make a new header that only contains celestial (i.e., on-sky) information
    new_hdr = WCS(header).celestial.to_header()
    new_hdr['NAXIS1'] = header['NAXIS1']
    new_hdr['NAXIS2'] = header['NAXIS2']
    return new_hdr


def master_mask(pcube):
    # create a 2D mask over where any of the paramater map has finite values
    mask = np.any(np.isfinite(pcube), axis=0)
    mask = mask_cleaning(mask)
    return mask


def mask_cleaning(mask):
    # designed to clean a noisy map, with a footprint that is likely slightly larger
    #mask = remove_small_objects(mask, min_size=9) #pending investigation before removed permantly
    mask = binary_dilation(mask, disk(1))
    mask = remove_small_holes(mask, 9)
    return mask


def regrid(image, header1, header2, dmask=None, method='cubic'):
    # similar to hcongrid from FITS_tools, but uses scipy.interpolate.griddata to interpolate over nan values
    grid1 = get_pixel_mapping(header1, header2)

    xline = np.arange(image.shape[1])
    yline = np.arange(image.shape[0])
    X,Y = np.meshgrid(xline, yline)

    mask = np.isfinite(image)

    if dmask is None:
        dmask = np.ones(grid1[0].shape, dtype=bool)

    return griddata((X[mask],Y[mask]), image[mask], (grid1[1]*dmask, grid1[0]*dmask), method=method, fill_value=np.nan)


def refine_guess(map, min=None, max=None, mask=None, disksize=1, scipy_interpolate=False):
    # refine parameter maps by outlier-fitering, masking, and interpolating
    map = map.copy()

    if min is not None:
        map[map<min] = np.nan
    if max is not None:
        map[map>max] = np.nan

    # check the number of finite pixels in the provided map
    mask_finite = np.isfinite(map)
    n_valid = np.sum(mask_finite)

    # in case there are too few valid pixels in the provided map
    if n_valid < 2:
        # if there's a single pixel
        if n_valid == 1:
            map[:] = map[mask_finite][0]

        # if there are no valid pixel in the guesses, set it to one of the limits or zero
        elif min is not None:
            if max is None:
                map[:] = min
            else:
                map[:] = (max + min) / 2
        elif max is not None:
            map[:] = max
        else:
            map[:] = 0.0
        return map

    if mask is None:
        mask = mask_finite
        mask = mask_cleaning(mask)

    def interpolate_scipy(map, mask):
        # interpolate over the dmask footprint
        xline = np.arange(map.shape[1])
        yline = np.arange(map.shape[0])
        X,Y = np.meshgrid(xline, yline)
        itpmask = np.isfinite(map)
        C = intp((X[itpmask],Y[itpmask]), map[itpmask])

        # interpolate over the dmask footprint
        zi = C(X*mask,Y*mask)
        return zi

    def interpolate_via_cnv(map, mask):
        mask_finite = np.isfinite(map)
        kernel = Gaussian2DKernel(2.5/2.355)
        zi = convolve(map, kernel, boundary='extend')
        # only populate pixels where the original map was finite
        zi[mask_finite] = map[mask_finite]

        # interpolate further to fill the mask if the mask is much larger than the kernel
        zi = interpolate.iter_expand(zi, mask=mask)

        return zi

    if scipy_interpolate:
        warn_msg = "The usage of scipy.interpolate for guess refine is deprecated and will be removed in a future version." \
                   "The default going forward will be astropy's convovle method."
        warnings.warn(warn_msg, DeprecationWarning, stacklevel=2)
        logger.warning(warn_msg)
        try:
            # interpolate the mask
            zi = interpolate_scipy(map, mask)
        except QhullError as e:
            logger.warning("qhull input error found; astropy convolve will be used instead")
            zi = interpolate_via_cnv(map, mask) # use astropy convolve as a proxy for interpolation
            
        except ValueError as e:
            logger.error("ValueError found (no points given); astropy convolve will be used instead")
            zi = interpolate_via_cnv(map, mask)
    else:
        zi = interpolate_via_cnv(map, mask)
    return zi

def refine_2c_guess(guesses, f_sigv = 0.5):
    # f_sigv is the fraction of sigma_v to increase the velocity seperation by

    v1, v2 = guesses[0], guesses[4]
    s1, s2 = guesses[1], guesses[5]

    vdiff = v1 - v2
    vdiff_sign = vdiff/np.abs(vdiff)
    sigvsum = s1 + s2

    # increase their v seperation by half of their total linewidth scaled by f_sigv
    v1 = v1 + sigvsum/2 * f_sigv * vdiff_sign
    v2 = v2 - sigvsum/2 * f_sigv * vdiff_sign

    # reduce the linewidth guesses, since they tend to be overestimated
    s1 = s1/2
    s2 = s2/2

    guesses[0], guesses[4] = v1, v2
    guesses[1], guesses[5] = s1, s2

    return guesses


def save_guesses(paracube, header, savename, ncomp=2):
    # a method to save the fitted parameter cube with relavent header information
    import copy
    from astropy.io import fits

    npara = 4
    hdr_new = copy.deepcopy(header)

    # write the header information for each plane (i.e., map)
    for i in range (0, ncomp):
        hdr_new['PLANE{0}'.format(i*npara+0)] = 'VELOCITY_{0}'.format(i+1)
        hdr_new['PLANE{0}'.format(i*npara+1)] = 'SIGMA_{0}'.format(i+1)
        hdr_new['PLANE{0}'.format(i*npara+2)] = 'TEX_{0}'.format(i+1)
        hdr_new['PLANE{0}'.format(i*npara+3)] = 'TAU_{0}'.format(i+1)

    hdr_new['CDELT3']= 1
    hdr_new['CTYPE3']= 'FITPAR'
    hdr_new['CRVAL3']= 0
    hdr_new['CRPIX3']= 1

    fitcubefile = fits.PrimaryHDU(data=paracube, header=hdr_new)
    fitcubefile.writeto(savename, overwrite=True)


import numpy as np
from astropy.io import fits
from plotly.subplots import make_subplots
import plotly.offline as pyo

from ..utils import dataframe as dframe
from ..moment_guess import peakT as quickPeakT

#======================================================================================================================#
from ..utils.mufasa_log import get_logger
logger = get_logger(__name__)


class ScatterPPV(object):
    """A class to plot the fitted parameters in 3D scatter plots. Most of the data is stored in a pandas DataFrame.

    Parameters
    ----------
    parafile : str
        Path to the .fits file containing the MUFASA generated parameter maps.
    fittype : str
        The name of the fit model, e.g., "nh3_multi_v" or "n2hp_multi_v".
    vrange : tuple of float, optional
        Velocity range to clip the data (in km/s). Data outside this range is excluded.
        Default is None.
    verr_thres : float, optional
        Velocity error threshold (in km/s) to filter out data with higher errors.
        Data with a velocity error greater than this threshold is excluded. Default is 5.

    Examples
    --------

    Initialize the ScatterPPV object and plot the position-position-velocity (PPV) scatter plot:

    >>> sc = scatter_3D.ScatterPPV("path/to/fname.fits", fittype="nh3_multi_v")
    >>> sc.plot_ppv(savename='monR2.html', vel_scale=0.5)
    """

    def __init__(self, parafile, fittype, vrange=None, verr_thres=5, meta_model=None):
        """Initialize the ScatterPPV object by loading data from a .fits file and setting up parameters.

        For a detailed description of parameters, refer to the class docstring.

        Parameters
        ----------
        parafile : str
            Path to the .fits file of the modeled parameter maps.
        fittype : str
            Name of the fit model.
        vrange : tuple of float, optional
            Velocity range to clip the data (in km/s).
        verr_thres : float, optional
            The velocity error threshold (in km/s) to filter the data. Data with errors above this threshold is excluded.
        """
        self.paracube, self.header = fits.getdata(parafile, header=True)
        self.fittype = fittype
        self.meta_model = meta_model

        if meta_model is None:
            # get the rest frequency
            if self.fittype == "nh3_multi_v":
                from pyspeckit.spectrum.models.ammonia_constants import freq_dict
                self.rest_freq = freq_dict['oneone']*1e-9 # in GHz

            elif self.fittype == "n2hp_multi_v":
                from ..spec_models.n2hp_constants import freq_dict
                self.rest_freq = freq_dict['onezero']*1e-9 # in GHz

        else:
            if fittype != meta_model.fittype:
                msg = f"The provided fittype ({fittype}) does not match that from the MetaModel ({meta_model.fittype})." \
                      f"MetaModel's fittype will be adopted over the provided on."
                logger.warning(msg)

            self.fittype = meta_model.fittype
            self.rest_freq = self.meta_model.rest_value.value

        # structure the data in the data frame
        self.dataframe = dframe.make_dataframe(self.paracube, vrange=vrange, verr_thres=verr_thres)

        # estimate the peak intensity of each spectral model
        self.add_peakI()

        # get the relative wcs coordinates
        self.add_wcs_del()

    def set_meta_model(self, meta_model):
        self.meta_model = meta_model


    def add_peakI(self):
        """Calculate and add a peak intensity value for each model point in the DataFrame.

        Parameters
        ----------
        nu : float, optional
            Reference frequency (in GHz) to estimate the peak intensity. If not provided, defaults to the `rest_freq` attribute.

        Returns
        -------
        None
        """
        if self.meta_model is not None:
            para = np.array(
                [self.dataframe['vlsr'].values,
                 self.dataframe['sigv'].values,
                 self.dataframe['tex'].values,
                 self.dataframe['tau'].values]
            )
            self.dataframe['peakT'] = self.meta_model.peakT(para)

        else:
            self.dataframe['peakT'] = quickPeakT(tex=self.dataframe['tex'], tau=self.dataframe['tau'], nu=self.rest_freq)


    def add_wcs_del(self, ra_ref=None, dec_ref=None, unit='arcmin'):
        """Calculate relative RA & Dec coordinates and add them to the DataFrame as columns.

        Parameters
        ----------
        ra_ref : float, optional
            Reference RA value to calculate relative RA. If not provided, uses the minimum RA in the data.
        dec_ref : float, optional
            Reference Dec value to calculate relative Dec. If not provided, uses the minimum Dec in the data.
        unit : {'arcmin', 'arcsec'}, optional
            Units for delta RA & Dec. Use 'arcmin' to plot in arcminutes or 'arcsec' for arcseconds. Default is 'arcmin'.

        Returns
        -------
        None
        """
        if unit == 'arcmin':
            f = 60
        elif unit == 'arcsec':
            f = 3600

        df = self.dataframe
        df['delt RA'] = df['x_crd'] * self.header['CDELT1'] * f * -1.0
        df['delt Dec'] = df['y_crd'] * self.header['CDELT2'] * f
        ''
        if ra_ref is None:
            ra_ref = df['delt RA'].min()
        if dec_ref is None:
            dec_ref = df['delt Dec'].min()

        df['delt RA'] = df['delt RA'] - ra_ref
        df['delt Dec'] = df['delt Dec'] - dec_ref

    def plot_ppv(self, label_key='peakT', vel_scale=0.8, xyunit='arcmin', savename=None, **kwargs):
        """
        Plot the fitted model in position-position-velocity (PPV) space as a 3D scatter plot.

        Points in the PPV plot are colored based on a specified key (e.g., intensity values)
        from the DataFrame. The velocity axis can be scaled relative to the spatial axes,
        and units for the x and y axes can be customized.

        Parameters
        ----------
        label_key : str, default='peakT'
            Column name in the DataFrame used for coloring data points.
            For example, 'peakT' for peak intensity or clustering labels.

        vel_scale : float, default=0.8
            Scaling factor for the velocity axis (z-axis) relative to the spatial axes.
            The x-axis is normalized to 1.

        xyunit : {'arcmin', 'pix'}, default='arcmin'
            Units for the x and y coordinates:
            - 'arcmin': Plot relative RA and Dec in arcminutes.
            - 'pix': Plot coordinates in pixels.

        savename : str, default=None
            File path to save the generated plot as an HTML file. If None, the plot is not saved.

        **kwargs : dict
            Additional keyword arguments passed to the `scatter_3D_df` function.
            Useful options include:

            auto_open_html : bool, default=True
                If True, automatically open the saved HTML file in a browser.

            mask_df : pandas.Series or None, default=None
                Boolean mask to filter the DataFrame before plotting.

        Returns
        -------
        fig : plotly.graph_objs.Figure
            A 3D scatter plot figure representing the PPV data.

        Other Parameters
        ----------------
        kwargs : dict
            Additional options for customizing the scatter plot, such as color mapping
            and opacity scaling.

        Notes
        -----
        - If `label_key` is "peakT", default color mapping ("magma_r") and opacity ranges
          are applied automatically.
        - The velocity axis range is derived from the 1st and 99th percentiles of the
          "vlsr" column, with additional padding.
        - Color scaling uses the 1st and 99th percentiles of the column specified by `label_key`.

        Examples
        --------
        >>> model.plot_ppv(label_key='peakT', vel_scale=0.9, xyunit='pix', savename='ppv_plot.html')
        >>> model.plot_ppv(label_key='cluster_label', auto_open_html=False, mask_df=mask)

        See Also
        --------
        scatter_3D_df : Function used internally for generating the scatter plot.
        """
        if label_key == 'peakT':
            kwdf = dict(cmap='magma_r', opacity_ranges=5)
            kwargs = {**kwdf, **kwargs}

        kwargs['label_key'] = label_key

        # get the velocity range to plot
        vmin, vmax = np.nanpercentile(self.dataframe['vlsr'], [1, 99])
        vpad = (vmax - vmin)/10
        if vpad > 0.5:
            vmax += vpad
        else:
            vmax += 0.5

        vmask = np.logical_and(self.dataframe['vlsr']>vmin, self.dataframe['vlsr']<vmax)

        if label_key is not None:
            # Calculate the 1st and 99th percentiles for color and opacity scaling
            cmin, cmax = np.nanpercentile(self.dataframe[label_key][vmask], [1, 99])
            kwargs['vmin'] = cmin
            kwargs['vmax'] = cmax

        if xyunit == 'arcmin':
            # plot delta RA & Dec in arcmin
            kwargs['x_key'] = 'delt RA'
            kwargs['y_key'] = 'delt Dec'
            kwargs['xlab'] = u'\u0394' + 'RA (arcmin)'
            kwargs['ylab'] = u"\u0394" + 'Dec (arcmin)'

        elif xyunit == 'pix':
            # plot x & y in pixel coordinates
            kwargs['x_key'] = 'y_crd'
            kwargs['y_key'] = 'delt Dec'
            kwargs['xlab'] = 'RA (pix)'
            kwargs['ylab'] = 'Dec (pix)'

        self.fig = scatter_3D_df(self.dataframe[vmask], z_key='vlsr',
                                 zlab='<i>v</i><sub>LSR</sub> (km s<sup>-1</sup>)',
                                 nx=self.header['NAXIS1'], ny=self.header['NAXIS2'],
                                 z_scale=vel_scale, savename=savename, **kwargs)

def scatter_3D_df(dataframe, x_key, y_key, z_key, label_key=None, mask_df=None,
                  auto_open_html=True, **kwargs):
    """A wrapper for scatter_3D to quickly plot a pandas DataFrame in 3D.

    Parameters
    ----------
    dataframe : pandas.DataFrame
        DataFrame containing the data to plot.
    x_key : str
        Column name for the x-axis values.
    y_key : str
        Column name for the y-axis values.
    z_key : str
        Column name for the z-axis values.
    label_key : str, optional
        Column name to color the data points by. If None, data points are plotted without color scaling. Default is None.
    mask_df : pandas.Series or None, optional
        Boolean mask to filter the DataFrame before plotting. Default is None.
    auto_open_html : bool, optional
        Whether to automatically open the HTML plot file upon saving. Default is True.
    kwargs : dict
        Additional keyword arguments for the 3D scatter plot.

    Returns
    -------
    fig : plotly.graph_objs.Figure
        The 3D scatter plot figure.
    """
    if mask_df is not None:
        dataframe = dataframe[mask_df]

    x, y, z = dataframe[x_key], dataframe[y_key], dataframe[z_key]

    if label_key is None:
        labels = None
    elif label_key in dataframe.keys():
        labels = dataframe[label_key]
    else:
        labels = label_key

    return scatter_3D(x, y, z, labels=labels, auto_open_html=auto_open_html, **kwargs)


def scatter_3D(x, y, z, labels=None, nx=None, ny=None, z_scale=0.8, shadow=True, fig=None, savename=None,
               scene=None, xlab=None, ylab=None, zlab=None, showfig=True, kw_line=None,
               cmap='Spectral_r', auto_open_html=True, vmin=None, vmax=None,
               opacity_ranges=1, **kwargs):
    """Plot a 3D scatter plot with optional opacity scaling for point ranges.

    Parameters
    ----------
    x : array-like
        X coordinates of the data points.
    y : array-like
        Y coordinates of the data points.
    z : array-like
        Z coordinates of the data points.
    labels : array-like, optional
        Label values for color scaling. If 'peakT' (default), opacity is split into ranges; otherwise, a single trace with full opacity is used.
    nx : int, optional
        Number of pixels in x to set the aspect ratio. Default is None.
    ny : int, optional
        Number of pixels in y to set the aspect ratio. Default is None.
    z_scale : float, optional
        Aspect ratio for z-axis relative to x and y axes. Default is 0.8.
    shadow : bool or float, optional
        Adds a shadow projection on the z-plane. Default is True.
    fig : plotly.graph_objs.Figure, optional
        Figure to plot on. If None, creates a new figure.
    savename : str, optional
        Path to save the plot as an HTML file. Default is None.
    scene : dict, optional
        Scene configuration for the 3D plot. Default is None.
    xlab : str, optional
        X-axis label. Default is None.
    ylab : str, optional
        Y-axis label. Default is None.
    zlab : str, optional
        Z-axis label. Default is None.
    showfig : bool, optional
        Whether to display the figure after plotting. Default is True.
    kw_line : dict, optional
        Dictionary of line properties for connecting points, if desired. Default is None.
    cmap : str, optional
        Colormap for data points. Default is 'Spectral_r'.
    auto_open_html : bool, optional
        If True, auto-opens saved HTML. Default is True.
    vmin : float, optional
        Minimum value for color scaling. Default is None.
    vmax : float, optional
        Maximum value for color scaling. Default is None.
    opacity_ranges : int, optional
        Number of opacity ranges (1 to 5). For 'peakT' labels, splits opacity into equal intervals over 1-99 percentile. Default is 1.
    kwargs : dict
        Additional keyword arguments for plotting.

    Returns
    -------
    fig : plotly.graph_objs.Figure
        Generated 3D scatter plot figure.
    """
    if fig is None:
        fig = make_subplots(rows=1, cols=1, specs=[[{'type': 'scatter3d'}]])

    if nx is None:
        nx = x.max()
    if ny is None:
        ny = y.max()

    # Define opacity levels and ranges if more than one range is specified
    opacity_levels = np.linspace(0.1, 1, opacity_ranges)
    if labels is not None and opacity_ranges > 1:
        # Calculate percentiles for the given number of opacity ranges
        percentiles = np.linspace(0, 100, opacity_ranges + 1)
        cutoffs = np.nanpercentile(labels, percentiles)

        # Plot each range with respective opacity
        for i, (opacity, cutoff_min, cutoff_max) in enumerate(zip(opacity_levels, cutoffs[:-1], cutoffs[1:])):
            mask = (labels >= cutoff_min) & (labels < cutoff_max)
            sub_x, sub_y, sub_z = x[mask], y[mask], z[mask]
            sub_labels = labels[mask]

            marker = dict(size=1, color=sub_labels, colorscale=cmap, opacity=opacity)
            if vmin is not None and vmax is not None:
                marker.update(cmin=vmin, cmax=vmax)

            kw_scatter3d = dict(mode='markers', marker=marker)

            if kw_line is not None:
                kw_line_default = dict(color=labels, width=2)
                line = {**kw_line_default, **kw_line}
                kw_scatter3d['line'] = line
                del kw_scatter3d['mode']

            fig.add_scatter3d(x=sub_x, y=sub_y, z=sub_z, mode='markers', marker=marker)
    else:
        # Single trace for non-peakT labels or opacity_ranges=1
        if labels is None:
            labels = '#1f77b4' #muted blue
        marker = dict(size=1, color=labels, colorscale=cmap, opacity=1.0)
        if vmin is not None and vmax is not None:
            marker.update(cmin=vmin, cmax=vmax)

        kw_scatter3d = dict(mode='markers', marker=marker)

        fig.add_scatter3d(x=x, y=y, z=z, mode='markers', marker=marker)

    # Configure the scene and other plot settings
    if scene is None:
        scene = dict(camera=dict(eye=dict(x=1.15, y=1.15, z=0.8)),
                     xaxis=dict(),
                     yaxis=dict(),
                     zaxis=dict(),
                     aspectmode='manual',
                     aspectratio=dict(x=1, y=1*ny/nx, z=z_scale)) #fixed aspect ratio

    fig.update_layout(scene=scene, showlegend=False)

    if xlab is not None:
        fig.update_layout(scene=dict(xaxis_title=xlab))
    if ylab is not None:
        fig.update_layout(scene=dict(yaxis_title=ylab))
    if zlab is not None:
        fig.update_layout(scene=dict(zaxis_title=zlab))

    if shadow:
        marker = dict(size=1, color=labels, colorscale=cmap, opacity=1.0)
        if vmin is not None and vmax is not None:
            marker.update(cmin=vmin, cmax=vmax)

        # display the shadow
        z_shadow = z.copy()
        if isinstance(shadow, bool):
            z_shadow[:] = np.nanmin(z) - 0.5

        elif isinstance(shadow, int) or isinstance(shadow, float):
            z_shadow[:] = shadow

        mk = {**marker, 'opacity':0.03}
        kw_scatter3d_mod = {**kw_scatter3d, "marker":mk}
        fig.add_scatter3d(x=x, y=y, z=z_shadow, **kw_scatter3d_mod)

        # add a low transparent layer of grey to make it look more like a "shadow"
        mk = {**marker, 'opacity': 0.02, 'color':'grey'}
        kw_scatter3d_mod = {**kw_scatter3d, "marker": mk}
        fig.add_scatter3d(x=x, y=y, z=z_shadow, **kw_scatter3d_mod)

    if showfig:
        # Set notebook mode to work in offline
        pyo.init_notebook_mode()
        fig.show()

    if savename is not None:
        fig.write_html(savename, auto_open=auto_open_html)

    return fig


"""
Utilities for working with .fits files, including functions adapted from
FITS_tools to reduce dependencies on the aging package.
"""

import numpy as np
import astropy.wcs as pywcs
from astropy import coordinates
from astropy import units as u

def downsample_header(header, factor, axis):
    """
    Downsample a FITS header along a specified axis.

    This function downsamples the provided FITS header along the specified axis
    using the FITS convention for axis numbering. The `CDELT` and `CRPIX`
    keywords are adjusted accordingly to reflect the downsampling.

    Parameters
    ----------
    header : astropy.io.fits.Header
        The FITS header to be downsampled. This is typically obtained from a
        FITS file.
    factor : float
        The downsampling factor. Must be greater than 0.
    axis : int
        The axis along which to apply the downsampling. Follows the FITS
        convention for axis numbering (1-based indexing).

    Returns
    -------
    astropy.io.fits.Header
        A copy of the FITS header with adjusted `CDELT` and `CRPIX` values to
        account for the downsampling.

    Notes
    -----
    - The function modifies a copy of the input header, leaving the original
      header unchanged.
    - The code here is borrowed from `FITS_tools.downsample`.

    Examples
    --------
    >>> from astropy.io import fits
    >>> header = fits.Header()
    >>> header['CDELT1'] = 0.1
    >>> header['CRPIX1'] = 50
    >>> downsampled_header = downsample_header(header, factor=2, axis=1)
    >>> print(downsampled_header['CDELT1'], downsampled_header['CRPIX1'])
    0.2 25.5
    """
    header = header.copy()

    cd = 'CDELT{0:d}'.format(axis)
    cp = 'CRPIX{0:d}'.format(axis)
    scalefactor = 1./factor
    header[cp] = (header[cp]-1)*scalefactor + scalefactor/2. + 0.5
    header[cd] = header[cd]*factor

    return header


def get_pixel_mapping(header1, header2):
    """
    Compute the pixel mapping between two FITS headers or WCS objects.

    This function determines how pixel coordinates in one FITS header
    or WCS (`header1`) map to pixel coordinates in another (`header2`).
    It returns a grid that describes the transformation in terms of
    world coordinates.

    Parameters
    ----------
    header1 : `~astropy.io.fits.Header` or `~astropy.wcs.WCS`
        The header or WCS corresponding to the input image whose pixel
        coordinates are being transformed.
    header2 : `~astropy.io.fits.Header` or `~astropy.wcs.WCS`
        The header or WCS corresponding to the output image onto which
        the input image will be interpolated.

    Returns
    -------
    grid : `~numpy.ndarray`
        A 2D array of shape `(2, n, m)` where `n` and `m` are the dimensions
        of `header2`. The array contains the mapped pixel coordinates from
        `header1` in the following order:
        - `grid[0, :, :]`: The y-coordinates in `header1`.
        - `grid[1, :, :]`: The x-coordinates in `header1`.

    Raises
    ------
    TypeError
        If either `header1` or `header2` is not an instance of
        `astropy.io.fits.Header` or `astropy.wcs.WCS`.
    NotImplementedError
        If the coordinate system type (`CTYPE`) in the headers is not
        recognized or if unit conversions between coordinate systems
        are not implemented.

    Notes
    -----
    - The function supports transformations between images that share
      compatible celestial coordinate systems, such as `GLON`/`GLAT`
      and `RA`/`DEC`.
    - If the coordinate systems differ, they will be transformed to
      align before computing the pixel mapping.
    - The code here is borrowed from `FITS_tools.downsample`.

    Examples
    --------
    >>> from astropy.io import fits
    >>> from astropy.wcs import WCS
    >>> header1 = fits.Header()  # Define input header or WCS
    >>> header2 = fits.Header()  # Define output header or WCS
    >>> grid = get_pixel_mapping(header1, header2)
    >>> print(grid.shape)
    (2, n, m)  # Example output shape corresponding to header2 dimensions
    """
    wcs1 = _load_wcs_from_header(header1)
    wcs2 = _load_wcs_from_header(header2)

    if not all([w1 == w2 for w1, w2 in zip(wcs1.wcs.ctype, wcs2.wcs.ctype)]):
        allowed_coords = ('GLON', 'GLAT', 'RA', 'DEC')
        if all([(any(word in w1 for word in allowed_coords) and
                 any(word in w2 for word in allowed_coords))
                for w1, w2 in zip(wcs1.wcs.ctype, wcs2.wcs.ctype)]):
            csys1 = _ctype_to_csys(wcs1.wcs)
            csys2 = _ctype_to_csys(wcs2.wcs)
            convert_coordinates = True
        else:
            raise NotImplementedError(
                "Unit conversions between {0} and {1} have not yet been implemented."
                .format(wcs1.wcs.ctype, wcs2.wcs.ctype)
            )
    else:
        convert_coordinates = False

    outshape = [wcs2.naxis2, wcs2.naxis1]
    yy2, xx2 = np.indices(outshape)

    lon2, lat2 = wcs2.wcs_pix2world(xx2, yy2, 0)

    if convert_coordinates:
        C2 = coordinates.SkyCoord(lon2, lat2, unit=(u.deg, u.deg), frame=csys2)
        C1 = C2.transform_to(csys1)
        lon2, lat2 = C1.spherical.lon.deg, C1.spherical.lat.deg

    xx1, yy1 = wcs1.wcs_world2pix(lon2, lat2, 0)
    grid = np.array([yy1.reshape(outshape), xx1.reshape(outshape)])

    return grid


def _load_wcs_from_header(header):
    """
    Load a WCS object from a FITS header or verify an existing WCS instance.

    This function ensures that a `pywcs.WCS` object is created from the provided
    FITS header or verifies that the input is already a valid WCS instance.
    It also adds `naxis1` and `naxis2` attributes to the WCS object if they
    are not already defined.

    Parameters
    ----------
    header : `~astropy.io.fits.Header` or `~pywcs.WCS`
        The FITS header or WCS object to process.

    Returns
    -------
    wcs : `~pywcs.WCS`
        A WCS object with required attributes set.

    Raises
    ------
    TypeError
        If the input is neither a valid FITS header nor a `pywcs.WCS` instance.

    Notes
    -----
    - The code here is borrowed from `FITS_tools.downsample`.
    """
    if issubclass(pywcs.WCS, header.__class__):
        wcs = header
    else:
        try:
            wcs = pywcs.WCS(header)
        except:
            raise TypeError("header must either be a astropy.io.fits.Header "
                            " or pywcs.WCS instance")

        if not hasattr(wcs, 'naxis1'):
            wcs.naxis1 = header['NAXIS1']
        if not hasattr(wcs, 'naxis2'):
            wcs.naxis2 = header['NAXIS2']

    return wcs


def _ctype_to_csys(wcs):
    """
    Convert WCS `CTYPE` information to a coordinate system name.

    This function determines the coordinate system (e.g., `fk5`, `fk4`, or `galactic`)
    based on the `CTYPE` attribute and the equinox in the provided WCS object.

    Parameters
    ----------
    wcs : `~pywcs.WCS`
        A WCS object containing `CTYPE` and `equinox` information.

    Returns
    -------
    csys : str
        The name of the coordinate system. Possible values are:
        - `'fk5'` for J2000 equinox.
        - `'fk4'` for B1950 equinox.
        - `'galactic'` for Galactic coordinates.

    Raises
    ------
    NotImplementedError
        If the equinox is neither 2000 nor 1950 when `CTYPE` indicates
        celestial coordinates.

    Notes
    -----
    - This code here is borrowed from `FITS_tools.downsample`.
    """
    ctype = wcs.ctype[0]
    if 'RA' in ctype or 'DEC' in ctype:
        if wcs.equinox == 2000:
            return 'fk5'
        elif wcs.equinox == 1950:
            return 'fk4'
        else:
            raise NotImplementedError("Non-fk4/fk5 equinoxes are not allowed")
    elif 'GLON' in ctype or 'GLAT' in ctype:
        return 'galactic'




"""
The `mufasa.moment_guess` module provides utilities for generating initial guesses, refining moment maps,
and handling physical parameter calculations such as excitation temperature and
optical depth.
"""

from __future__ import print_function
import numpy as np
import copy

from astropy import units as u
from spectral_cube import SpectralCube

import pyspeckit
from pyspeckit.spectrum.models.ammonia_constants import (ckms, h, kb)
from astropy.stats import mad_std

from .utils import map_divide
import multiprocessing
from .utils.multicore import validate_n_cores
#=======================================================================================================================
from .utils.mufasa_log import get_logger
logger = get_logger(__name__)
#=======================================================================================================================


class LineSetup(object):
    def __init__(self, linetype='nh3', new_recipe=True):

        if linetype == 'nh3':
            from pyspeckit.spectrum.models.ammonia_constants import voff_lines_dict

            voff = voff_lines_dict['oneone']

            # define max and min values of tex and tau to use for the test
            # a spectrum with tex and tau values both below the specified minima has an intensity below the expected GAS rms

            if new_recipe:
                self.tex_max = 8.0
                self.tau_max = 5.0
                self.tex_min = 3.0
                self.tau_min = 0.1
            else:
                self.tex_max = 8.0
                self.tau_max = 1.0
                self.tex_min = 3.1
                self.tau_min = 0.3

        elif linetype == 'n2hp':
            from .spec_models.n2hp_constants import voff_lines_dict

            voff = voff_lines_dict['onezero']

            # define max and min values of tex and tau to use for the test
            if new_recipe:
                self.tex_max = 8.0
                self.tau_max = 5.0
                self.tex_min = 3.0
                self.tau_min = 0.1
            else:
                self.tex_max = 8.0
                self.tau_max = 1.0
                self.tex_min = 3.1
                self.tau_min = 0.3
        else:
            raise Exception("{} is an invalid linetype".format(linetype))

#=======================================================================================================================

def master_guess(spectrum, ncomp, sigmin=0.07, v_peak_hwidth=3.0, v_atpeak=None, widewVSep=False, snr_cut=3,
                 signal_mask=None, linetype='nh3'):

    m0, m1, m2 = window_moments(spectrum, window_hwidth=v_peak_hwidth, v_atpeak=v_atpeak, signal_mask=signal_mask)

    # estimate the rms level, and pass to the spectrum (probably should've been performed in another function)
    rms = get_rms_prefit(spectrum, window_hwidth=v_peak_hwidth, v_atpeak=m1)

    if m0 < snr_cut*rms:
        gg = np.zeros((ncomp * 4,) + np.array([m1], dtype=object).shape)
        gg[:] = np.nan
        return gg

    if ncomp == 2 and widewVSep:
        # use recipe that recovers two-slab spectra (warning, may not be ideal if more than 3 slabs are present)

        m0_b, m1_b, m2_b = noisemask_moment(spectrum, m1, m2, mask_sigma=4, noise_rms=rms, window_hwidth=v_peak_hwidth)

        if m0_b > snr_cut*rms:
            # if the residual spectrum has m0 that is 3 sigma above the rms noise, treat both moment as individual
            # one component parts
            gg_a = moment_guesses(np.array([m1]), np.array([m2]), ncomp=1, sigmin=sigmin, moment0=np.array([m0]),
                                  linetype=linetype)
            gg_b = moment_guesses(np.array([m1_b]), np.array([m2_b]), ncomp=1, sigmin=sigmin, moment0=np.array([m0_b]),
                                  linetype=linetype)
            gg = np.zeros((ncomp * 4,) + np.array([m1]).shape)
            gg[:4,:] = gg_a[:]
            gg[4:,:] = gg_b[:]

        else:
            gg = moment_guesses(np.array([m1]), np.array([m2]), ncomp, sigmin=sigmin, moment0=np.array([m0]), linetype=linetype)

    else:
        # get the guesses based on moment maps based on "traditional" recipe
        gg = moment_guesses(np.array([m1]), np.array([m2]), ncomp, sigmin=sigmin, moment0=np.array([m0]), linetype=linetype)

    return gg


# THIS FUNCTION IS NEVER  CALLED
def get_window_slab(maskcube, window_hwidth=3.0, v_atpeak=None):
    if v_atpeak is None:
        # find the peak of the integrated spectrum if v_atpeak isn't provided
        tot_spec = np.nansum(maskcube._data[:,]*maskcube.get_mask_array(), axis=(1,2))
        idx_peak = np.nanargmax(tot_spec)
        print("peak T_B: {0}".format(np.nanmax(tot_spec)))
        v_atpeak = maskcube.spectral_axis[idx_peak].to(u.km/u.s).value
        print("v_atpeak: {0}".format(v_atpeak))

    vmax = v_atpeak + window_hwidth
    vmin = v_atpeak - window_hwidth

    # Extract the spectrum within the window defined around the main hyperfine components and take moments
    slab = maskcube.spectral_slab(vmin*u.km/u.s, vmax*u.km/u.s)
    return slab


def vmask_moments(cube, vmap, window_hwidth=3.0):
    # obtain moments with windows centered around the vlsr specified in the provided map
    cubemasked = vmask_cube(cube, vmap, window_hwidth)

    m0 = cubemasked.moment0(axis=0).value
    m1 = cubemasked.moment1(axis=0).to(u.km/u.s).value
    m2 = (np.abs(cubemasked.moment2(axis=0))**0.5).to(u.km/u.s).value

    return m0, m1, m2


def vmask_cube(cube, vmap, window_hwidth=3.0):
    spax = cube.spectral_axis.value
    spax_cube = np.ones(cube.shape) * spax[:, None, None]
    v_up = vmap + window_hwidth
    v_down = vmap - window_hwidth
    mask = np.logical_and(spax_cube > v_down, spax_cube < v_up)
    cubemasked = cube.with_mask(mask)
    return cubemasked


def window_moments(spec, window_hwidth=4.0, v_atpeak=None, signal_mask=None):
    """
    Calculate the zeroth, first, and second moments of a spectrum or cube
    within a specified velocity window.

    Parameters
    ----------
    spec : pyspeckit.Cube, pyspeckit.spectrum.classes.Spectrum, or SpectralCube
        The spectrum or cube for which moments are calculated.
        Can be a `pyspeckit.Cube`, `pyspeckit.spectrum.classes.Spectrum`,
        or `SpectralCube` object.
    window_hwidth : float, optional
        The half-width of the spectral window in km/s used to calculate moments.
        This is useful for isolating hyperfine lines. Default is 4.0.
    v_atpeak : float, numpy.ndarray, or None, optional
        The velocity or velocity map (in km/s) around which to center the moment
        calculation. If None, it will be estimated from the spectrum or cube.
        Default is None.
    signal_mask : numpy.ndarray or None, optional
        An optional mask to apply when estimating the velocity peak (`v_atpeak`).
        Useful for suppressing noise when determining the peak. Default is None.

    Returns
    -------
    tuple of numpy.ndarray
        A tuple containing:
        - Zeroth moment (`m0`) as a numpy array.
        - First moment (`m1`) as a numpy array in km/s.
        - Second moment (`m2`) as a numpy array in km/s.

    Raises
    ------
    Exception
        If the input `spec` is not a supported type.

    Notes
    -----
    - This function acts as a wrapper to calculate moments for different types
      of inputs (`SpectralCube`, `pyspeckit.Cube`, or `pyspeckit.spectrum`).
    - For `SpectralCube`, `v_atpeak` as a map is not currently supported.

    Examples
    --------
    For a `SpectralCube` object:

    >>> from spectral_cube import SpectralCube
    >>> cube = SpectralCube.read("example_cube.fits")
    >>> m0, m1, m2 = window_moments(cube, window_hwidth=3.0, v_atpeak=5.0)

    For a `pyspeckit.spectrum.classes.Spectrum`:

    >>> import pyspeckit
    >>> spec = pyspeckit.Spectrum("example_spectrum.fits")
    >>> m0, m1, m2 = window_moments(spec, window_hwidth=2.0)
    """
    def moments_pys_spectrum(spectrum, window_hwidth=4.0, v_atpeak=None, iter_refine=False):
        '''
        find moments within a given window (e.g., around the main hyperfine lines)
        # note: iter_refine has not proven to be very effective in our tests

        :param spectrum:
            <pyspeckit.spectrum.classes.Spectrum>
            the spectrum to take the momentw of

        :param window_hwidth: float
            half-width of the window (in km/s) to be used to isolate the main hyperfine lines from the rest of the spectrum

        '''

        if v_atpeak is None:
            moments = spectrum.moments(unit=u.km/u.s)
            v_atpeak = moments[2]

        vmax = v_atpeak + window_hwidth
        vmin = v_atpeak - window_hwidth

        # Extract the spectrum within the window defined around the main hyperfine components and take moments
        slice = spectrum.slice(vmin, vmax, unit=u.km/u.s)
        moments = slice.moments(unit=u.km/u.s)

        if iter_refine:
            # for low snr- this method really doesn't work well
            m0, m1, m2 = moments[1], moments[2], moments[3]
            # make the window smaller by making out channels outside a specific width around moment 1
            # create a window 2 times the second moment
            new_window_hw = m2*3.0
            if new_window_hw > window_hwidth:
                new_window_hw = window_hwidth
            vmax = m1 + new_window_hw
            vmin = m1 - new_window_hw
            slice = spectrum.slice(vmin, vmax, unit=u.km / u.s)
            moments = slice.moments(unit=u.km / u.s)

        return moments[1], moments[2], moments[3]

    def moments_pys_cube(pcube, window_hwidth=4.0, v_atpeak=None, iter_refine=False, multicore=None):
        """
        Calculate moments of a pyspeckit spectral cube within a specified velocity window.

        Parameters
        ----------
        pcube : pyspeckit.cubes.SpectralCube.Cube
            The spectral cube for which moments are calculated.
        window_hwidth : float, optional
            The half-width of the velocity window (in km/s) used to isolate spectral lines.
            Default is 4.0 km/s.
        v_atpeak : float, numpy.ndarray, or None, optional
            The velocity or velocity map (in km/s) around which to center the moment
            calculations. If None, moments are calculated over the entire cube and
            refined iteratively. Default is None.
        iter_refine : bool, optional
            If True, refines the velocity window iteratively based on the first moment.
            This is useful for low signal-to-noise spectra. Default is False.
        multicore : int or None, optional
            The number of CPU cores to use for parallel processing. If None, it will use
            all available cores. Default is None.

        Returns
        -------
        tuple of numpy.ndarray
            A tuple containing:
            - Zeroth moment (`m0`) as a numpy array.
            - First moment (`m1`) as a numpy array in km/s.
            - Second moment (`m2`) as a numpy array in km/s.

        Notes
        -----
        - This function uses pyspeckit's `momenteach` method for calculating moments.
        - If `v_atpeak` is provided, the function will restrict calculations to the
          specified velocity window. Otherwise, an initial pass over the cube will
          estimate the velocity map, which is used to define the window.
        - Iterative refinement of moments can be computationally expensive and
          is not recommended for large datasets unless necessary.

        Examples
        --------
        >>> from pyspeckit import Cube
        >>> pcube = Cube("example_cube.fits")
        >>> m0, m1, m2 = moments_pys_cube(pcube, window_hwidth=3.0)

        >>> vmap = pcube.moment1(unit='km/s').value
        >>> m0, m1, m2 = moments_pys_cube(pcube, window_hwidth=2.0, v_atpeak=vmap, iter_refine=True)

        Raises
        ------
        AttributeError
            If parallel processing with `momenteach` fails, the function will fall back
            to single-core execution.
        """
        multicore = validate_n_cores(multicore)

        def get_win_moms(pcube, v_atpeak):
            # get window moments when v_atpeak is given
            pcube_masked = window_mask_pcube(pcube, v_atpeak, win_hwidth=window_hwidth)
            try:
                pcube_masked.momenteach(unit='km/s', verbose=False, multicore=multicore)
            except AttributeError:
                # if multicore fails, only use one core
                pcube_masked.momenteach(unit='km/s', verbose=False, multicore=1)
            moments = pcube_masked.momentcube
            return moments

        # note: the current implementation may be memory intensitive
        if v_atpeak is None:
            # note the moment 1 estimate of pcube.momenteach seem to be pretty good at ignoring hyperfine structures
            try:
                pcube.momenteach(unit='km/s', vheight=False, verbose=False, multicore=multicore)
            except AttributeError:
                # if multicore fails, only use one core
                pcube.momenteach(unit='km/s', vheight=False, verbose=False, multicore=1)

            moments = pcube.momentcube
            # redo again with the hyperfine lines masked out
            moments = get_win_moms(pcube, v_atpeak=moments[1])
        else:
            moments = get_win_moms(pcube, v_atpeak)

        if iter_refine:
            # use the moment 1 estimated velocity to define new windows
            # for low snr this method may not work well
            moments = get_win_moms(pcube, v_atpeak=moments[1])

        return moments[0], moments[1], moments[2]


    def moments_spectralcube(maskcube, window_hwidth, v_atpeak=None, signal_mask=None):
        # signal_mask is to provide additional masking specifically and only for v_atpeak estimate
        if v_atpeak is None:
            # find the peak of the integrated spectrum if v_atpeak isn't provided
            mask = maskcube.get_mask_array()
            if signal_mask is not None:
                mask = mask*signal_mask
            tot_spec = np.nansum(maskcube._data[:,]*mask, axis=(1,2))
            idx_peak = np.nanargmax(tot_spec)
            logger.debug("Getting window moments of SpectralCube")
            logger.debug("peak T_B: {0}".format(np.nanmax(tot_spec)))
            v_atpeak = maskcube.spectral_axis[idx_peak].to(u.km/u.s).value
            logger.debug("v_atpeak: {0}".format(v_atpeak))

        vmax = v_atpeak + window_hwidth
        vmin = v_atpeak - window_hwidth

        # Extract the spectrum within the window defined around the main hyperfine components and take moments
        slab = maskcube.spectral_slab(vmin*u.km/u.s, vmax*u.km/u.s)
        m0 = slab.moment0(axis=0).value
        m1 = slab.moment1(axis=0).to(u.km/u.s).value
        m2 = (np.abs(slab.moment2(axis=0))**0.5).to(u.km/u.s).value

        return m0, m1, m2
    
    # wrapper to find moments for different types of inputs

    if isinstance(spec, pyspeckit.Cube):
        # this method is much slower than using SpectralCube, but also seems more robust at spectral peaks
        return moments_pys_cube(spec, window_hwidth, v_atpeak)

    elif isinstance(spec, pyspeckit.spectrum.classes.Spectrum):
        return moments_pys_spectrum(spec, window_hwidth, v_atpeak)

    elif isinstance(spec, SpectralCube):
        # currently cannot handle v_atpeak as a map
        if hasattr(v_atpeak, 'ndim') and v_atpeak.ndim>0: # numpy floats have ndim=0
            logger.error("the method that handles SpectralCube cannot currently handle v_atpeak as a map, please use single value v_atpeak instead")
        return moments_spectralcube(spec, window_hwidth, v_atpeak, signal_mask)

    else:
        raise Exception("the spec provided is of invalid type")
        return None


def noisemask_moment(sp, m1, m2, mask_sigma=4, noise_rms = None, **kwargs):
    # mask out the 'main' component based on moment map and replace them with fake noise
    # and rerun window_mements to find additional components

    sp_m = copy.copy(sp)

    if 'v_atpeak' not in kwargs:
        kwargs['v_atpeak'] = m1

    if noise_rms is None:
        noise_rms = get_rms_prefit(sp, **kwargs)

    mask = np.logical_and(sp_m.xarr.value < m1 + mask_sigma * m2, sp_m.xarr.value > m1 - mask_sigma * m2)
    sp_m.data[mask] = np.random.randn(np.sum(mask)) * noise_rms

    return window_moments(sp_m, **kwargs)


def moment_guesses(moment1, moment2, ncomp, sigmin=0.07, tex_guess=3.2, tau_guess=0.5, moment0=None, linetype='nh3', mom0_floor=None):
    """
    Generate reasonable initial guesses for multiple component fits based on moment maps.

    Parameters
    ----------
    moment1 : numpy.ndarray
        The first moment (velocity centroid) map.
    moment2 : numpy.ndarray
        The second moment (velocity dispersion) map.
    ncomp : int
        Number of components for the fit.
    sigmin : float, optional
        Minimum velocity dispersion (in km/s). Default is 0.07 km/s, the spectral resolution of the GAS channels.
    tex_guess : float, optional
        Initial guess for the excitation temperature (T_ex). Default is 3.2 K.
    tau_guess : float, optional
        Initial guess for the optical depth (tau). Default is 0.5.
    moment0 : numpy.ndarray, optional
        Zeroth moment (integrated intensity) map. If provided, this will be used to
        modify the guesses for T_ex and tau. Default is None.
    linetype : str, optional
        Line type, either 'nh3' or 'n2hp'. Determines parameter limits. Default is 'nh3'.
    mom0_floor : float, optional
        Minimum floor value for moment0 when normalizing. If not provided, a default
        normalization is applied. Default is None.

    Returns
    -------
    numpy.ndarray
        A 2D array of shape `(ncomp * 4, moment1.shape)` containing the guesses for each parameter:
        - Component velocity centroids.
        - Velocity dispersions.
        - Excitation temperatures.
        - Optical depths.

    Notes
    -----
    - For `ncomp == 1`, the method uses a single-component guess based on the provided moments.
    - For `ncomp == 2`, the method applies a recipe for a bright and faint component with
      offsets in velocity and reduced T_ex and tau for the faint component.
    - For `ncomp > 2`, components are evenly spaced in velocity, and parameter guesses
      are scaled accordingly.
    - If `moment0` is provided, the guesses for T_ex and tau are scaled based on the
      normalized moment0 values.

    Examples
    --------
    >>> import numpy as np
    >>> moment1 = np.array([[1.0, 2.0], [3.0, 4.0]])
    >>> moment2 = np.array([[0.1, 0.2], [0.3, 0.4]])
    >>> ncomp = 2
    >>> guesses = moment_guesses(moment1, moment2, ncomp, tex_guess=4.0, tau_guess=0.8)

    >>> moment0 = np.array([[1.5, 2.5], [3.5, 4.5]])
    >>> guesses_with_mom0 = moment_guesses(moment1, moment2, ncomp, moment0=moment0, mom0_floor=0.1)

    Raises
    ------
    ValueError
        If `ncomp` is less than 1 or other parameters are invalid.
    """
    # setup different limits based on the line specie
    line_setup = LineSetup(linetype)
    tex_max = line_setup.tex_max
    tau_max = line_setup.tau_max
    tex_min = line_setup.tex_min
    tau_min = line_setup.tau_min

    if moment0 is not None:
        #print "[WARNING]: moment0 map is provided, thus the user-provided tex and tau will not be used"
        # normalize the moment 0 map with respect to the norm_ref percentile value
        # e.g., 95 percentile value being normalized to have a value of 0.95
        norm_ref = 99.73
        mom0high = np.percentile(moment0[np.isfinite(moment0)], norm_ref)
        # may want to modify this normalization to be something a little simpler or physical (i.e., 99.73/100 ~ 1)
        if mom0_floor is None:
            m0Norm = moment0.copy()*norm_ref/100.0/mom0high
            tex_guess = m0Norm*tex_max
            tau_guess = m0Norm*tau_max
        else:
            m0Norm = moment0.copy() * norm_ref / 100.0 / (mom0high - mom0_floor)
            tex_guess = m0Norm * (tex_max - tex_min) + tex_min
            tau_guess = m0Norm * (tau_max - tau_min) + tau_min

    m1 = moment1
    m2 = moment2

    # Guess linewidth (the current recipe works okay, but potential improvements can be made.
    gs_sig = m2/ncomp
    gs_sig[gs_sig < sigmin] = sigmin
    # note 0.08 k is narrow enough to be purely thermal @ ~10 K

    # there are 4 parameters for each v-component
    gg = np.zeros((ncomp*4,)+m1.shape)

    if ncomp == 1:
        gg[0,:] = m1                 # v0 centriod
        gg[1,:] = gs_sig             # v0 width
        gg[2,:] = tex_guess          # v0 T_ex
        gg[3,:] = tau_guess          # v0 tau

    # using a working recipe (assuming a bright and a faint componet)
    if ncomp == 2:
        sigmaoff = 0.4
        tau2_frac = 0.25                    # the tau weight of the second component relative to the total fraction
        gg[0,:] = m1 - sigmaoff*m2         # v0 centriod
        gg[1,:] = gs_sig                   # v0 width
        gg[2,:] = tex_guess                # v0 T_ex
        gg[3,:] = tau_guess*(1-tau2_frac)  # v0 tau
        gg[4,:] = m1 + sigmaoff*m2         # v1 centriod
        gg[5,:] = gs_sig                   # v1 width
        gg[6,:] = tex_guess*0.8            # v1 T_ex
        gg[7,:] = tau_guess*tau2_frac      # v1 tau

    # using a generalized receipe that have not been tested (lots of room for improvement!)
    if ncomp > 2:
        for i in range (0, ncomp):
            gg[i,  :] = m1+(-1.0+i*1.0/ncomp)*0.5*m2 # v0 centriod (step through a range fo velocities within sigma_v)
            gg[i+1,:] = gs_sig                   # v0 width
            gg[i+2,:] = tex_guess*0.8            # v0 T_ex
            gg[i+3,:] = tau_guess/ncomp*0.25     # v0 tau


    # ensure the tex and tau guesses falls within the guessing limits
    tex_guess[tex_guess < tex_min] = tex_min
    tex_guess[tex_guess > tex_max] = tex_max
    tau_guess[tau_guess < tau_min] = tau_min
    tau_guess[tau_guess > tau_max] = tau_max

    return gg

#=======================================================================================================================
# additional recipe

def moment_guesses_1c(m0, m1, m2):
    # make guesses based on the moment maps
    # note: it does not impose limits on parameters such as tau and tex

    # m0 from pyspeckit is actually amplitube proxy, so to calculate
    gs_sig = m2 ** 0.5
    mom0 = m0 * gs_sig * np.sqrt(2 * np.pi)

    mom0_thres = 3.0 # the regime boundary between fix-tau- or fix-tex-based calculation
    mom0_min = 0.03 # Ta ~ 0.25 at Tex=3, tau=0.1
    tau_fx = 2.5 #1.5 #2.5
    tex_fx = 6.0 #7.0 #6.0 # K

    # divide the tau tex guess into two regimes by integrated flux
    # if flux is greater than mom0_thres, assume a fixed tau value and caculate the corrosponding tex assume Gaussian
    # otherwise, assume a fixed tex value and calculate tau instead
    if isinstance(m0, float):
        # for 0D
        if mom0 < mom0_min:
            mom0 = mom0_min

        if mom0 > mom0_thres:
            tau_guess = tau_fx
            tex_guess = get_tex(mom0, tau_fx)

        else:
            tex_guess = tex_fx
            tau_guess = get_tau(mom0, tex_guess)

    elif m0.ndim == 2:
        # aassume all mom0 lower than mom0 to be mom0_min
        mom0[mom0 < mom0_min] = mom0_min

        # for 2D
        tau_guess = np.zeros(m0.shape)
        tex_guess = np.zeros(m0.shape)
        mask = mom0 > mom0_thres

        tau_guess[mask] = tau_fx
        tex_guess[mask] = get_tex(mom0[mask], tau_fx)

        tau_guess[~mask] = get_tau(mom0[~mask], tex_fx)
        tex_guess[~mask] = tex_fx


    else:
        raise Exception("the moment 0 input has the wrong dimension ({})".format(m0.ndim))
        return None

    return np.asarray([m1, gs_sig, tex_guess, tau_guess])


def mom_guess_wide_sep(spec, vpeak=None, rms=None, planemask=None, multicore=None):
    # for two components

    win_hwidth = 4.0
    # the window for the second component recovery (though the 1st win_hwidth should mask out the hyperfines already)
    win_hwidth2 = 10.0

    # the amount of moment 2 estimated sigma to maskout for "residual moment maps"
    f_sig_mask = 2.0

    rms_thres = 2.0  # the amplitude threshold to count as the signal

    f_tau = 0.5 # weight of the tau for the "equal-weight" guesses
    f_sig = 0.5 # weight of the linewidth for all guesses

    multicore = validate_n_cores(multicore)

    if isinstance(spec, pyspeckit.spectrum.classes.Spectrum):
        spec.xarr.velocity_convention = 'radio'
        spec.xarr = spec.xarr.as_unit('km/s')

        if rms is None:
            # we only need a crude estimate
            rms = mad_std(spec.data, axis=None, ignore_nan=True)

        if vpeak is None:
            # find the v_peak pased on smoothed spectrum
            sp_smooth = spec.copy()
            sp_smooth.smooth(3)
            idx = np.argmax(sp_smooth.data)
            vpeak = sp_smooth.xarr[idx]
            vpeak = vpeak.value
            logger.info("vpeak for widely separated moment guesses: {}".format(vpeak))

        # get the moments
        m0, m1, m2 = window_moments(spec, window_hwidth=win_hwidth, v_atpeak=vpeak)

        # assume the emission is dominated by the brighter component, and use moment maps to
        # create guesses for the first component
        gg1 = moment_guesses_1c(m0, m1, m2)

        # convert moment 2 to sigma
        sig = m2 ** 0.5

        # mask emission found by moment masks, and try to find a secondary peak using a broader window
        sp4 = spec.copy()
        mask = np.logical_and(sp4.xarr.value > m1 - sig * f_sig_mask, sp4.xarr.value < m1 + sig * f_sig_mask)
        #mask2 = np.logical_and(sp4.xarr.value > vpeak - win_hwidth, sp4.xarr.value < vpeak + win_hwidth)
        mask2 = np.logical_and(sp4.xarr.value > m1 - win_hwidth, sp4.xarr.value < m1 + win_hwidth)
        mask = np.logical_or(mask, ~mask2)

        sp4.data[mask] = 0.0
        #sp4.data[~mask2] = 0.0

        m0n, m1n, m2n = window_moments(sp4, window_hwidth=win_hwidth2, v_atpeak=vpeak)

        gg2 = moment_guesses_1c(m0n, m1n, m2n)
        gg = np.concatenate((gg1, gg2))

        if m0n < rms * rms_thres:
            gg[4:8] = gg1[:]
            gg[0] += sig
            gg[4] -= sig
            # set tau of each component to half of what the 1-component moment guess is
            gg[3] *= f_tau
            gg[7] *= f_tau

        gg[1] *= f_sig
        gg[5] *= f_sig


    elif isinstance(spec, SpectralCube):

        if rms is None:
            # we only need a crude estimate
            rms = mad_std(spec._data, axis=0, ignore_nan=True)

        spec = spec.with_spectral_unit("km/s", velocity_convention="radio")
        pcube = pyspeckit.Cube(cube=spec, maskmap=planemask, velocity_convention = 'radio')

        # get the moments using the pyspeckit method
        m0, m1, m2 = window_moments(pcube, window_hwidth=win_hwidth, v_atpeak=vpeak)

        # assume the emission is dominated by the brighter component, and use moment maps to
        # create guesses for the first component
        gg1 = moment_guesses_1c(m0, m1, m2)

        # convert moment 2 to sigma
        sig = m2 ** 0.5

        # maskout emission found by moment masks, and try to find a secondary peak
        spax = spec.spectral_axis.value
        spax_cube = np.broadcast_to(spax[:, np.newaxis, np.newaxis], (len(spax), m1.shape[0], m1.shape[1]))
        smask = np.logical_and(spax_cube > m1 - sig * f_sig_mask, spax_cube < m1 + sig * f_sig_mask)
        smask2 = np.logical_and(spax_cube > m1 - win_hwidth, spax_cube < m1 + win_hwidth)
        smask = np.logical_or(smask, ~smask2)

        maskcube = spec.with_mask(~smask)
        # fill value needs to be zero for moment estimates to work as expected
        maskcube = maskcube.with_fill_value(0.0)

        # convert it to pyspeckit to take advantage of pyspeckit's moment methods
        pcube = pyspeckit.Cube(cube=maskcube, maskmap=planemask)
        pcube.momenteach(verbose=False, multicore=multicore)
        m0n, m1n, m2n = pcube.momentcube

        gg2 = moment_guesses_1c(m0n, m1n, m2n)

        # use "equal weight guesses" if the remaining spectrum is too faint
        mask = m0n < rms * rms_thres

        gg2[:,mask] = gg1[:,mask].copy()
        gg = np.concatenate((gg1, gg2), axis=0)

        #gg[4:8, mask] = gg1[:, mask]
        gg[0, mask] += sig[mask]
        gg[4, mask] -= sig[mask]
        # set tau of each component to half of what the 1-component moment guess is
        gg[3,mask] *= f_tau
        gg[7,mask] *= f_tau
        # set the guessing linewidth to be half of the moment guesses, for both components

        gg[1,:] *= f_sig
        gg[5,:] *= f_sig


    return gg


#=======================================================================================================================
# utility functions

def get_rms_prefit(spectrum, window_hwidth, v_atpeak, linetype='nh3'):

    s = spectrum
    line_setup = LineSetup(linetype)

    vsys = v_atpeak*u.km/u.s
    throw = window_hwidth*u.km/u.s
    voff = line_setup.voff

    mask = np.ones(s.shape[0], dtype=np.bool)

    for deltav in voff:
        mask *= (np.abs(s.xarr - (deltav * u.km / u.s + vsys)) > throw)

    d_rms = s.data.copy()

    return mad_std(d_rms[mask])


def adaptive_moment_maps(maskcube, seeds, window_hwidth, weights=None, signal_mask=None):
    # split the cube up into different regions and make mosaic moment maps from moments of each individual regions

    labels, n_labs = map_divide.dist_divide(seeds, weights=weights, return_nmarkers=True)

    if signal_mask is None:
        signal_mask = seeds

    m0 = np.zeros(labels.shape)
    m0[:] = np.nan
    m1 = m0.copy()
    m2 = m0.copy()

    moments = [m0, m1, m2]

    # make moment map for each region
    for i in range(n_labs):
        mask = labels == i + 1

        new_cube_mask = maskcube.get_mask_array()
        new_cube_mask = new_cube_mask * mask

        moms = window_moments(maskcube.with_mask(new_cube_mask), window_hwidth, signal_mask=signal_mask*mask)

        for m, m_p in zip(moments, moms):
            m[mask] = m_p[mask]

    return m0, m1, m2


def window_mask_pcube(pcube, vmid, win_hwidth=4.0):
    # returns a copy of the pucbe (pyspeckit.cubes.SpectralCube.Cube) with spectra outside of the window zeroed
    shape = pcube.cube.shape
    spax = pcube.xarr.value
    spax_cube = np.broadcast_to(spax[:, np.newaxis, np.newaxis], (len(spax), shape[1], shape[2]))

    smask = np.logical_and(spax_cube > vmid - win_hwidth, spax_cube < vmid + win_hwidth)

    pcube = pcube.copy()
    #pcube.cube[~smask] = 0.0
    pcube.cube[~smask] = np.nan
    pcube.maskmap = np.any(np.isfinite(pcube.cube), axis=0)

    return pcube

#=======================================================================================================================
# physics functions

def get_tex(Ta, tau=0.5, nu=23.722634):
    # calculate the excitation temperature given tau

    background_tb = 2.7315
    T0 = (h * nu * 1e9 / kb)

    term1 = Ta / T0 / (1 - np.exp(-tau))
    term2 = 1.0 / (np.exp(T0 / background_tb) - 1)
    term3 = 1.0 / (term1 + term2) + 1
    term4 = T0 / np.log(term3)
    return term4


def get_tau(Ta, tex=6.0, nu=23.722634):
    background_tb = 2.7315
    T0 = (h * nu * 1e9 / kb)

    term1 = 1 / (np.exp(T0 / tex) - 1) - 1 / (np.exp(T0 / background_tb) - 1)
    term2 = -Ta / T0 / term1 + 1
    return -np.log(term2)


def peakT(tex, tau, nu=23.722634):
    background_tb = 2.7315
    tauprof = tau
    T0 = (h * nu * 1e9 / kb)
    return (T0 / (np.exp(T0 / tex) - 1) - T0 / (np.exp(T0 / background_tb) - 1)) * (1 - np.exp(-tauprof))



"""
The `mufasa.deblend_cube` module provides functionality for deblending hyperfine structures in spectral cubes,
allowing for the reconstruction of fitted models with Gaussian lines accounting for
optical depths.
"""

from __future__ import print_function
from __future__ import absolute_import
__author__ = 'mcychen'

#=======================================================================================================================

# import external library
import numpy as np
from spectral_cube import SpectralCube
from astropy.utils.console import ProgressBar
import astropy.units as u
import gc

from pyspeckit.parallel_map import parallel_map

from .utils.multicore import validate_n_cores

#=======================================================================================================================

def deblend(para, specCubeRef, vmin=4.0, vmax=11.0, f_spcsamp=None, tau_wgt=0.1, n_cpu=None, linetype='nh3'):
    """
    Deblend hyperfine structures in a cube based on fitted models.

    This function reconstructs the fitted model with Gaussian lines accounting
    for optical depths (e.g., similar to CO (J = 0-1)).

    Parameters
    ----------
    para : numpy.ndarray
        The fitted parameters in the order of velocity, width, Tex, and tau for
        each velocity slab. The size of the z-axis for `para` must be a multiple of 4.
    specCubeRef : spectral_cube.SpectralCube
        The reference cube from which the deblended cube is constructed.
    vmin : float, optional
        The lower velocity limit on the deblended cube in km/s. Default is 4.0.
    vmax : float, optional
        The upper velocity limit on the deblended cube in km/s. Default is 11.0.
    f_spcsamp : int, optional
        The scaling factor for spectral sampling relative to the reference cube.
        For example, `f_spcsamp=2` doubles the spectral resolution.
    tau_wgt : float, optional
        A scaling factor for the input tau. For example, `tau_wgt=0.1` better
        represents the true optical depth of an NH3 (1,1) hyperfine group than
        the "fitted tau". Default is 0.1.
    n_cpu : int, optional
        The number of CPUs to use. If None, defaults to all CPUs available minus one.
    linetype : str, optional
        The line type to use for deblending. Options are 'nh3' or 'n2hp'. Default is 'nh3'.

    Returns
    -------
    mcube : spectral_cube.SpectralCube
        The deblended cube.
    """

    # get different types of deblending models
    if linetype == 'nh3':
        from .spec_models import nh3_deblended
        deblend_mod = nh3_deblended.nh3_vtau_singlemodel_deblended

    elif linetype == 'n2hp':
        from .spec_models import n2hp_deblended
        deblend_mod = n2hp_deblended.n2hp_vtau_singlemodel_deblended
    else:
        raise Exception("{} is an invalid linetype".format(linetype))

    # open the reference cube file
    cube = specCubeRef
    cube = cube.with_spectral_unit(u.km/u.s, velocity_convention='radio')

    # trim the cube to the specified velocity range
    cube = cube.spectral_slab(vmin*u.km/u.s,vmax*u.km/u.s)

    # generate an empty SpectralCube to house the deblended cube
    if f_spcsamp is None:
        deblend = np.zeros(cube.shape)
        hdr = cube.wcs.to_header()
        wcs_new = cube.wcs
    else:
        deblend = np.zeros((cube.shape[0]*int(f_spcsamp), cube.shape[1], cube.shape[2]))
        wcs_new = cube.wcs.deepcopy()
        # adjust the spectral reference value
        wcs_new.wcs.crpix[2] = wcs_new.wcs.crpix[2]*int(f_spcsamp)
        # adjust the spaxel size
        wcs_new.wcs.cdelt[2] = wcs_new.wcs.cdelt[2]/int(f_spcsamp)
        hdr = wcs_new.to_header()

    # retain the beam information
    hdr['BMAJ'] = cube.header['BMAJ']
    hdr['BMIN'] = cube.header['BMIN']
    hdr['BPA'] = cube.header['BPA']

    mcube = SpectralCube(deblend, wcs_new, header=hdr)

    # convert back to an unit that the ammonia hf model can handle (i.e. Hz) without having to create a
    # pyspeckit.spectrum.units.SpectroscopicAxis object (which runs rather slow for model building in comparison)
    mcube = mcube.with_spectral_unit(u.Hz, velocity_convention='radio')
    xarr = mcube.spectral_axis

    yy,xx = np.indices(para.shape[1:])
    # a pixel is valid as long as it has a single finite value
    isvalid = np.any(np.isfinite(para),axis=0)
    valid_pixels = zip(xx[isvalid], yy[isvalid])

    def model_a_pixel(xy):
        x,y = int(xy[0]), int(xy[1])
        # nh3_vtau_singlemodel_deblended takes Hz as the spectral unit
        models = [deblend_mod(xarr, Tex=tex, tau=tau*tau_wgt, xoff_v=vel, width=width)
                  for vel, width, tex, tau in zip(para[::4, y,x], para[1::4, y,x], para[2::4, y,x], para[3::4, y,x])]

        mcube._data[:,y,x] = np.nansum(np.array(models), axis=0)
        return ((x, y), mcube._data[:, y, x])

    n_cpu = validate_n_cores(n_cpu)

    if n_cpu > 1:
        print("------------------ deblending cube -----------------")
        print("number of cpu used: {}".format(n_cpu))
        sequence = [(x, y) for x, y in valid_pixels]
        result = parallel_map(model_a_pixel, sequence, numcores=n_cpu)
        merged_result = [core_result for core_result in result
                         if core_result is not None]
        for mr in merged_result:
            ((x, y), model) = mr
            x = int(x)
            y = int(y)
            mcube._data[:, y, x] = model
    else:
        for xy in ProgressBar(list(valid_pixels)):
            model_a_pixel(xy)

    # convert back to km/s in units before saving
    mcube = mcube.with_spectral_unit(u.km/u.s, velocity_convention='radio')
    gc.collect()
    print("--------------- deblending completed ---------------")

    return mcube



"""
The `mufasa.aic` module provides tools for statistical evaluation of spectral
cube models, including Akaike Information Criterion (AIC) calculations, corrected
AIC (AICc), and chi-squared metrics.

"""
from __future__ import absolute_import
__author__ = 'mcychen'

import numpy as np
import astropy.io.fits as fits
from spectral_cube import SpectralCube

from . import multi_v_fit as mvf

#=======================================================================================================================

def fits_comp_AICc(cubepath, modpath1, modpath2, aiccpath, likelihoodpath=None):
    """
    A wrapper function to calculate corrected Akaike Information Criterion (AICc) values
    and save them to a FITS file.

    Parameters
    ----------
    cubepath : str
        Path to the data cube FITS file.
    modpath1 : str
        Path to the first model FITS file.
    modpath2 : str
        Path to the second model FITS file.
    aiccpath : str
        Path to save the resulting AICc FITS file.
    likelihoodpath : str, optional
        Path to save the likelihood FITS file, if provided.

    Returns
    -------
    None
    """

    cube = SpectralCube.read(cubepath)
    mod1, hdr1 = fits.getdata(modpath1, header = True)
    mod2, hdr2 = fits.getdata(modpath2, header = True)

    aicc1, aicc2 = get_comp_AICc(cube, mod1, mod2, p1 = 4, p2 = 8)

    hdr_new = cube.wcs.celestial.to_header()
    hdr_new['PLANE1'] = "AICc values for the 1 component fit model"
    hdr_new['PLANE2'] = "AICc values for the 2 component fit model"

    aicccube = fits.PrimaryHDU(data=np.array([aicc1, aicc2]), header=hdr_new)
    aicccube.writeto(aiccpath, overwrite=True)

    if likelihoodpath is not None:
        likelyhood = (aicc1 - aicc2) / 2.0
        fits.writeto(likelihoodpath, likelyhood, cube.wcs.celestial.to_header(), overwrite=True)


def fits_comp_chisq(cubepath, modpath1, modpath2, savepath, reduced=True):
    """
    Calculate and save chi-squared values for the given cube and model fits.

    Parameters
    ----------
    cubepath : str
        Path to the data cube FITS file.
    modpath1 : str
        Path to the first model FITS file.
    modpath2 : str
        Path to the second model FITS file.
    savepath : str
        Path to save the resulting chi-squared FITS file.
    reduced : bool, optional
        Whether to calculate reduced chi-squared values. Default is True.

    Returns
    -------
    None
    """
    cube = SpectralCube.read(cubepath)
    mod1, hdr1 = fits.getdata(modpath1, header = True)
    mod2, hdr2 = fits.getdata(modpath2, header = True)

    hdr_new = cube.wcs.celestial.to_header()
    hdr_new['PLANE1'] = "reduced chi-squared values for the 1 component fit model"
    hdr_new['PLANE2'] = "reduced chi-squared values for the 2 component fit model"

    mask1 = mod1 > 0
    mask2 = mod2 > 0
    mask = np.logical_or(mask1, mask2)

    # expand of 20 is same as that used to calculate aic value
    chi1 = mvf.get_chisq(cube, mod1, expand=20, reduced = reduced, usemask = True, mask = mask)
    chi2 = mvf.get_chisq(cube, mod2, expand=20, reduced = reduced, usemask = True, mask = mask)

    chicube = fits.PrimaryHDU(data=np.array([chi1, chi2]), header=cube.wcs.celestial.to_header())
    chicube.writeto(savepath, overwrite=True)


def get_comp_AICc(cube, model1, model2, p1, p2):
    """
    Calculate AICc values for two models over the same samples.

    Parameters
    ----------
    cube : SpectralCube
        The data cube.
    model1 : numpy.ndarray
        The first model cube.
    model2 : numpy.ndarray
        The second model cube.
    p1 : int
        Number of parameters associated with the first model.
    p2 : int
        Number of parameters associated with the second model.

    Returns
    -------
    tuple of numpy.ndarray
        AICc values for the first and second models.
    """
    mask1 = model1 > 0
    mask2 = model2 > 0
    mask = np.logical_or(mask1, mask2)

    chi1, N1 = mvf.get_chisq(cube, model1, expand=20, reduced = False, usemask = True, mask = mask)
    chi2, N2 = mvf.get_chisq(cube, model2, expand=20, reduced = False, usemask = True, mask = mask)

    # I need a way to double check that N1 and N2 are the same (just in case)
    aicc1 = AICc(chi1, p1, N1)
    aicc2 = AICc(chi2, p2, N1)

    return aicc1, aicc2


def AIC(rss, p, N):
    """
    Calculate the Akaike Information Criterion (AIC).

    Parameters
    ----------
    rss : numpy.ndarray
        Residual sum of squares.
    p : int
        Number of parameters.
    N : int
        Number of samples.

    Returns
    -------
    numpy.ndarray
        AIC values.
    """
    # avoid invalid math values
    N[N==0] = np.nan
    aic = N * np.log(rss/N) + 2*p
    #return np.nan_to_num(aic)
    return aic


def AICc(rss, p, N):
    """
    Calculate the corrected Akaike Information Criterion (AICc).

    Parameters
    ----------
    rss : numpy.ndarray
        Residual sum of squares.
    p : int
        Number of parameters.
    N : int
        Number of samples.

    Returns
    -------
    numpy.ndarray
        Corrected AICc values.
    """
    top = 2*p*(p+1)
    bottom = N - p - 1
    return AIC(rss, p, N) + top/bottom


def likelihood(aiccA, aiccB):
    """
    Calculate the log-likelihood of model A relative to model B.

    Parameters
    ----------
    aiccA : numpy.ndarray
        AICc values for model A.
    aiccB : numpy.ndarray
        AICc values for model B.

    Returns
    -------
    numpy.ndarray
        Log-likelihood values.
    """
    #aiccA, aiccB = np.nan_to_num(aiccA), np.nan_to_num(aiccB)
    return -1.0*(aiccA - aiccB) / 2.0





"""
The `mufasa.convolve_tools` module provides tools for processing spectral cubes with spatial
convolution, signal-to-noise masking, and edge trimming.
"""

from __future__ import print_function
__author__ = 'mcychen'

import numpy as np
import astropy.io.fits as fits
from astropy import units as u
from skimage.morphology import remove_small_objects, disk, opening, binary_erosion, dilation, remove_small_holes
from spectral_cube import SpectralCube
from radio_beam import Beam
from astropy.wcs import WCS
from astropy.stats import mad_std
from astropy.convolution import Gaussian2DKernel, convolve
from scipy.interpolate import griddata
import scipy.ndimage as nd
from spectral_cube.utils import NoBeamError
import gc

# for Astropy 6.1.4 forward compatibility
try:
    from astropy.units import UnitScaleError
except ImportError:
    from astropy.units.core import UnitScaleError

from .utils.fits_utils import downsample_header, get_pixel_mapping

#=======================================================================================================================
from .utils.mufasa_log import get_logger
logger = get_logger(__name__)
#=======================================================================================================================
# utility tools for convolve cubes

def convolve_sky_byfactor(cube, factor, savename=None, edgetrim_width=5, downsample=True, **kwargs):
    # factor = factor * 1.0 # probably unecessary, better option is factor = float(factor)

    if not isinstance(cube, SpectralCube):
        cube = SpectralCube.read(cube, use_dask=True)

    if edgetrim_width is not None:
        cube = edge_trim(cube, trim_width=edgetrim_width)

    hdr = cube.header

    # sanity check
    if hdr['CUNIT1'] != hdr['CUNIT2']:
        raise Exception("the spatial axis units for the cube do not match each other!")
        return None

    beamunit = getattr(u, hdr['CUNIT1'])
    bmaj = hdr['BMAJ'] * beamunit * factor
    bmin = hdr['BMIN'] * beamunit * factor
    pa = hdr['BPA']

    try:
        beam = Beam(major=bmaj, minor=bmin, pa=pa)
    except UnitScaleError:
        beam = Beam(major=bmaj, minor=bmin, pa=None)

    # convolve
    try:
        # for Astropy 6.1.4 forward compatibility
        cnv_cube = convolve_sky(cube, beam, **kwargs)
    except NoBeamError:
        cube = cube.with_beam(beam)
        cnv_cube = convolve_sky(cube, beam, **kwargs)

    if not np.isnan(cnv_cube.fill_value):
        cnv_cube = cnv_cube.with_fill_value(np.nan)

    if downsample:
        # regrid the convolved cube
        nhdr = downsample_header(hdr, factor=factor, axis=1)
        nhdr = downsample_header(nhdr, factor=factor, axis=2)
        nhdr['NAXIS1'] = int(np.rint(hdr['NAXIS1']/factor))
        nhdr['NAXIS2'] = int(np.rint(hdr['NAXIS2']/factor))
        newcube = cnv_cube.reproject(nhdr, order='bilinear')
    else:
        newcube = cnv_cube

    if savename != None:
        newcube.write(savename, overwrite=True)

    return newcube


def convolve_sky(cube, beam, snrmasked=False, iterrefine=False, snr_min=3.0):
    # return the convolved cube in the same gridding as the input
    # note: iterrefine masks data as well

    if not isinstance(cube, SpectralCube):
        cube = SpectralCube.read(cube, use_dask=True)

    if not np.isnan(cube.fill_value):
        cube = cube.with_fill_value(np.nan)

    mask = np.any(cube.mask.include(), axis=0)

    if snrmasked:
        planemask = snr_mask(cube, snr_min)
        plane_mask_size = np.sum(planemask)
        if plane_mask_size > 25:
            mask = mask & planemask
            logger.info("snr plane mask size = {}".format(plane_mask_size))
        else:
            logger.warning("snr plane mask too small (size = {}), no snr mask is applied".format(plane_mask_size))

    maskcube = cube.with_mask(mask)

    # enable huge operations (https://spectral-cube.readthedocs.io/en/latest/big_data.html for details)
    if maskcube.size > 1e8:
        logger.warning("maskcube is large ({} pixels)".format(maskcube.size))
    maskcube.allow_huge_operations = True
    cnv_cube = maskcube.convolve_to(beam)
    maskcube.allow_huge_operations = False
    gc.collect()

    if snrmasked and iterrefine:
        # use the convolved cube for new masking
        logger.debug("--- second iteration refinement ---")
        mask = cube.mask.include()
        planemask = snr_mask(cnv_cube, snr_min)
        plane_mask_size = np.sum(planemask)
        if np.sum(planemask) > 25:
             mask = mask*planemask
             logger.info("snr plane mask size = {}".format(plane_mask_size))
        else:
            logger.warning("snr plane mask too small (size = {}), no snr mask is applied".format(plane_mask_size))
        maskcube = cube.with_mask(mask)
        maskcube.allow_huge_operations = True
        cnv_cube = maskcube.convolve_to(beam)
        maskcube.allow_huge_operations = False
        gc.collect()

    return cnv_cube


def snr_mask(cube, snr_min=1.0, errmappath=None):
    # create a mask around the cube with a snr cut

    if errmappath is not None:
        errmap = fits.getdata(errmappath)

    else:
        # make a quick RMS estimate using median absolute deviation (MAD)
        errmap = cube.mad_std(axis=0)#, how='ray')
        logger.info("median rms: {0}".format(np.nanmedian(errmap)))

    snr = cube.filled_data[:].value / errmap
    peaksnr = np.nanmax(snr, axis=0)

    #the snr map will inetiabley be noisy, so a little smoothing
    kernel = Gaussian2DKernel(1)
    peaksnr = convolve(peaksnr, kernel)

    def default_masking(snr, snr_min):
        planemask = (snr > snr_min)

        if planemask.size > 100:
            # attempt to remove noisy features
            planemask = binary_erosion(planemask, disk(1))
            planemask_im = remove_small_objects(planemask, min_size=9)
            if np.sum(planemask_im) > 9:
                # only adopt the erroded mask if there are objects left in it
                planemask = planemask_im
            # note, dialation is larger than erosion so the foot print is a bit more extended
            planemask = dilation(planemask, disk(3))

        return (planemask)

    planemask = default_masking(peaksnr, snr_min)
    del peaksnr  # Free memory
    gc.collect()

    return planemask


def edge_trim(cube, trim_width=3):
        # trim the edges by N pixels to guess the location of the peak emission
        mask = np.any(cube.mask.include(), axis=0)
        #mask = np.any(np.isfinite(cube._data), axis=0)
        if mask.size > 100:
            mask = binary_erosion(mask, disk(trim_width))
        mask = cube.mask.include() & mask

        return cube.with_mask(mask)


def regrid_mask(mask, header, header_targ, tightBin=True):
    # calculate scaling ratio between the two images
    yratio = np.abs(header['CDELT2']/header_targ['CDELT2'])
    xratio = np.abs(header['CDELT2']/header_targ['CDELT2'])
    maxratio = np.max([yratio, xratio])

    if (maxratio <= 0.5) & tightBin:
        # erode the mask a bit to avoid binning artifacts when downsampling
        s = 2
        kern = np.ones((s, s), dtype=bool)
        mask = binary_erosion(mask, structure=kern)

    # using the fits convention of x and y
    shape = (header_targ['NAXIS2'], header_targ['NAXIS1'])

    # regrid a boolean mask
    grid = get_pixel_mapping(header_targ, header)

    if (maxratio <= 0.5):
        # the mapping seems a little off for the y-axis when downsampling
        # works for factor of 2 grid, but may want to check and see if this is an issue with any relative pixel size grid
        grid[0] = grid[0] + 1.0
        outbd = grid[0]> shape[0]
        # make sure the coordinates are not out of bound
        grid[0][outbd] = grid[0][outbd] - 1.0

    grid = grid.astype(int)

    newmask = np.zeros(shape, dtype=bool)
    newmask[grid[0, mask], grid[1, mask]] = True

    if maxratio > 1:
        # dilate the mask to preserve the footprint
        s = int(maxratio - np.finfo(np.float32).eps) + 1
        kern = np.ones((s+1,s+1), dtype=bool)
        kern[-1,:] = False
        kern[:,0] = False
        newmask = nd.binary_dilation(newmask, structure=kern)

    return newmask


def regrid(image, header1, header2, dmask=None, method='cubic'):
    # similar to hcongrid from FITS_tools, but uses scipy.interpolate.griddata to interpolate over nan values
    grid1 = get_pixel_mapping(header1, header2)

    xline = np.arange(image.shape[1])
    yline = np.arange(image.shape[0])
    X,Y = np.meshgrid(xline, yline)

    mask = np.isfinite(image)

    if dmask is None:
        dmask = np.ones(grid1[0].shape, dtype=bool)

    return griddata((X[mask],Y[mask]), image[mask], (grid1[1]*dmask, grid1[0]*dmask), method=method, fill_value=np.nan)


def get_celestial_hdr(header):
    # make a new header that only contains celestial (i.e., on-sky) information
    new_hdr = WCS(header).celestial.to_header()
    new_hdr['NAXIS1'] = header['NAXIS1']
    new_hdr['NAXIS2'] = header['NAXIS2']
    return new_hdr






import numpy as np
from skimage.morphology import disk, footprint_rectangle

def maxref_neighbor_coords(mask, ref, fill_coord=(0, 0), structure=None, centre=None):
    # find pixel of a neighbour with the highest reference value
    highest_coords = []

    # Get coordinates of mask
    true_pixel_coords = np.argwhere(mask)

    if structure is None:
        structure, centre = square_neighbour(r=1, return_centre=True)
    else:
        if isinstance(structure, str):
            if structure == '4-connect':
                # 4 nearest neighbours
                structure, centre = disk_neighbour(1, return_centre=True)
            elif structure == '8-connect':
                # 8 nearest neighbours
                structure, centre = square_neighbour(r=1, return_centre=True)

    if centre is None:
        centre = structure.shape
        centre = ( np.ceil(centre[0]/2).astype(int) - 1 , np.ceil(centre[1]/2).astype(int) - 1 )

    # For each pixel within the mask, find its valid neighbors
    for y, x in true_pixel_coords:
        neighbors = get_valid_neighbors(ref, x, y, structure=structure, centre=centre)

        if neighbors:
            # Find the pixel with the highest reference value among the neighbors
            highest_neighbor = max(neighbors, key=lambda x: x[1])  # Find the highest neighbor
            highest_coord = highest_neighbor[0]  # Get the coordinates of the highest ref pixel
        else:
            highest_coord = fill_coord

        highest_coords.append(highest_coord)

    return highest_coords

def get_valid_neighbors(data, x, y, structure, centre):
    neighbors = []

    ycoord, xcoord = get_neighbor_coord(structure, x,y, centre[1], centre[0])

    for y, x in zip(ycoord, xcoord):
        try:
            neighbors.append(((y, x), data[y, x]))
        except IndexError:
            continue

    return neighbors

def get_neighbor_coord(struct, x, y, x_ref, y_ref):
    idx = np.where(struct)
    ycoord = idx[0] - y_ref + y
    xcoord = idx[1] - x_ref + x
    return (ycoord, xcoord)

def disk_neighbour(r, return_centre=False):
    struct = disk(r, dtype='bool')
    struct[r, r] = False
    if return_centre:
        return struct, (r,r)
    else:
        return struct

def square_neighbour(r, return_centre=False):
    # r is the degree of seperation from
    width = r*2+1
    struct = footprint_rectangle((width, width), dtype='bool')
    struct[r, r] = False
    if return_centre:
        return struct, (r,r)
    else:
        return struct


import numpy as np
from astropy.io import fits
import pandas as pd

def read(filename, header=True, vrange=None, verr_thres=5):
    """
    Read a FITS file and convert the data to a pandas DataFrame, optionally including the header.

    Parameters
    ----------
    filename : str
        Path to the .fits file containing the MUFASA output parameter data.
    header : bool, optional
        Whether to include the header in the output. Default is True.
    vrange : tuple of float, optional
        Velocity range to filter the data, specified as (v_min, v_max) in km/s. Default is None, which includes all data.
    verr_thres : float, optional
        Velocity error threshold (in km/s) to exclude data points with higher errors. Default is 5.

    Returns
    -------
    dataframe : pandas.DataFrame
        A DataFrame containing the parameter data, including columns such as 'vlsr', 'x_crd', 'y_crd', 'eVlsr', 'sigv',
         'tex', 'tau', and 'comp_i'.
    hdr : astropy.io.fits.Header, optional
        The header from the FITS file, included if `header` is True.

    Notes
    -----
    The function reads a FITS file, converts the parameter data into a structured DataFrame, and applies the specified
     velocity and error thresholds if provided.
    """
    para, hdr = fits.getdata(filename, header=header)
    dataframe = make_dataframe(para, vrange, verr_thres=verr_thres)

    if header:
        return dataframe, hdr
    else:
        return dataframe


def make_dataframe(para_2c, vrange=None, verr_thres=5):
    """
    Create a DataFrame from a 3D parameter array, applying optional velocity and error thresholds.

    Parameters
    ----------
    para_2c : numpy.ndarray
        A 3D array containing parameter values across two spatial dimensions and one parameter axis.
    vrange : tuple of float, optional
        Velocity range to filter data, specified as (v_min, v_max) in km/s. Default is None, which includes all data.
    verr_thres : float, optional
        Velocity error threshold (in km/s) to exclude data points with higher errors. Default is 5.

    Returns
    -------
    dataframe : pandas.DataFrame
        A DataFrame containing extracted and filtered data, with columns such as 'vlsr', 'x_crd', 'y_crd', 'eVlsr',
        'sigv', 'tex', 'tau', and 'comp_i'.

    Notes
    -----
    The function extracts relevant parameters from `para_2c`, applies the specified velocity and error thresholds,
    and structures the data into a pandas DataFrame.
    """

    ncomp = int(para_2c.shape[0] / 8)

    if vrange is None:
        v_min, v_max = None, None
    else:
        v_min, v_max = vrange

    if v_min is None:
        v_min = -1.0 * np.inf
    if v_max is None:
        v_max = np.inf

    # get the coordinate grid
    nz, ny, nx = para_2c.shape
    x_crds, y_crds = np.meshgrid(range(nx), range(ny), indexing='xy')

    data = {
        'x_crd': [], 'y_crd': [],
        'vlsr': [], 'sigv': [], 'tex': [], 'tau': [],
        'eVlsr': [], 'eSigv': [], 'eTex': [], 'eTau':[],
        'comp_i': []
    }

    for i in range(ncomp):
        # loop through the two components
        mask = np.isfinite(para_2c[i])

        # impose an error threshold
        mask = np.logical_and(mask, para_2c[i * 4 + 8] < verr_thres)

        # impose velocity range threshold
        mask = np.logical_and(mask, para_2c[i * 4] < v_max)
        mask = np.logical_and(mask, para_2c[i * 4] > v_min)

        data['x_crd'] += x_crds[mask].tolist()
        data['y_crd'] += y_crds[mask].tolist()

        data['vlsr'] += para_2c[i * 4][mask].tolist()
        data['sigv'] += para_2c[i * 4 + 1][mask].tolist()
        data['tex'] += para_2c[i * 4 + 2][mask].tolist()
        data['tau'] += para_2c[i * 4 + 3][mask].tolist()

        data['eVlsr'] += para_2c[i * 4 + 8][mask].tolist()
        data['eSigv'] += para_2c[i * 4 + 9][mask].tolist()
        data['eTex'] += para_2c[i * 4 + 10][mask].tolist()
        data['eTau'] += para_2c[i * 4 + 11][mask].tolist()

        data['comp_i'] += [i] * np.sum(mask) # comp_i is there to break the xy-degeneracy when adding more info

    return pd.DataFrame(data)


def assign_to_dataframe(dataframe, new_map, comp_i):
    """
    Assign values from a new data array to an existing DataFrame based on spatial coordinates and component index.

    Parameters
    ----------
    dataframe : pandas.DataFrame
        The main DataFrame containing spatial and component information.
    new_map : numpy.ndarray
        Array of new values to be assigned to the DataFrame, with shape (j, k).
    comp_i : int
        The component index associated with the new values.

    Returns
    -------
    dataframe : pandas.DataFrame
        Updated DataFrame with the new values assigned in a 'new_value' column.

    Notes
    -----
    This function merges the new values with the input DataFrame based on the x and y spatial coordinates and `comp_i`.
    """
    # Generate x and y coordinate grid for the new_map array (shape (j, k))
    j, k = new_map.shape
    x_crds, y_crds = np.meshgrid(range(k), range(j), indexing='xy')

    # Flatten the new_map and coordinate grids to create a DataFrame with the constant comp_i
    new_data = pd.DataFrame({
        'x_crd': x_crds.flatten(),
        'y_crd': y_crds.flatten(),
        'comp_i': comp_i,  # Add the constant comp_i to match with the main dataframe
        'new_value': new_map.flatten()
    })

    # Merge the new values based on 'x_crd', 'y_crd', and 'comp_i' with the input dataframe
    dataframe = dataframe.merge(new_data, on=['x_crd', 'y_crd', 'comp_i'], how='left')

    return dataframe
